--[[

Copyright (c) 2012-2013 gao7.com

http://www.gao7.com/LizardMan/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

]]

--[[--

横排数字类，定义横排数字相关操作方法及逻辑实现。

-   存储标签[跑马灯]相关信息。
-   示例：
    -- 引用类
    local LinearNum = require("cardui.LinearNum")

    local node = LinearNum.new({
        resource               = PHI_FIGHT_NUM .. "f_num",               -- 资源名(不包含后缀名的前缀，如要缓存要引入num.plist和num.png，则resource为num)
        prefix                 = "f_num_a",                             -- 数字及符号前缀
        value                  = 300,                                   -- 值
        isShowPlus             = true,                                  -- 显示＋标识
        isShowMinus            = true,                                  -- 显示－标识
        zeroSymbol             = true,                                  -- 数字0显示符号 可选值：+或-
        margin                 = ccp(0, 0),                             -- 项目间隔
        marginSymbol           = ccp(10, 0),
        align                  = LinearNum.ALIGN_CENTER,                -- 项目对齐方式
    })
    self:addChild(node, 99999)
    node:setPosition(ccp(display.cx, display.cy))

    local nodeCenter = display.newSprite(PHI_FIGHT .. "f_cemetery_board.png")
    self:addChild(nodeCenter, 9998)
    nodeCenter:setPosition(ccp(display.cx, display.cy))    
]]

local M = {}



-- [包含文件]
-- ..


-- [公共参数]
-- 常量
-- 默认间隔
M.DFT_MARGIN                    = ccp(0, 0)


-- 操作变量
-- 验证
local TDAssertUtils             = require(PH_MDL_UTIL .. "TDAssertUtils")
-- 字符串
local TDStringUtils             = require(PH_MDL_UTIL .. "TDStringUtils")
-- 数组
local TDArrayUtils              = require(PH_MDL_UTIL .. "TDArrayUtils")


-- 枚举
-- 符号
M.SYMBOL_POSITIVE               = 1
M.SYMBOL_NEGATIVE               = 2
M.SYMBOL_NONE                   = 3
M.SYMBOLS = {
    [1] = "0+",
    [2] = "0-",
    [3] = "0_",
}
-- 排列方式
M.ALIGN_CENTER                  = 1
M.ALIGN_LEFT                    = 2
M.ALIGN_RIGHT                   = 3
M.ALIGN_NONE                    = 4
-- 类型
M.KIND_SYMBOL                   = 1
M.KIND_NUM                      = 2










-- [构造方法]
--[[--
    创建对象
]]
function M.new(params)
    -- [属性定义]
    local _resource             -- 资源名(不包含后缀名的前缀，如要缓存要引入num.plist和num.png，则_resource为num)
    local _prefix               -- 数字及符号前缀
    local _value                -- 值
    local _node                 -- 结点
    local _resourcePNG          -- 资源图片文件.png
    local _resourcePList        -- 资源信息文件.plist
    local _charArray            -- 字符数组
    local _isShowPlus           -- 显示＋标识
    local _isShowMinus          -- 显示－标识
    local _frameNamePlus        -- 帧名称[＋]
    local _frameNameMinus       -- 帧名称[－]
    local _frameName0Symbol     -- 帧名称[数字0显示符号]
    local _margin               -- 项目间隔
    local _marginNum            -- 数字间隔
    local _marginSymbol         -- 符号间隔
    local _items                -- 项目集


    -- [私有方法定义]
    -- 获得帧名称
    local function getFrameName(tag)
        local frameName

        if type(tag) == "number" then 
            frameName = "#" .. _prefix .. tostring(tag) .. ".png"
        elseif type(tag) == "string" then
            if tag == "+" then 
                frameName = "#" .. _prefix .. "plus.png"
            elseif tag == "-" then
                frameName = "#" .. _prefix .. "minus.png"
            elseif tag == "0+" then
                frameName = "#" .. _prefix .. "plus.png"
            elseif tag == "0-" then
                frameName = "#" .. _prefix .. "minus.png"
            elseif tag == "0_" then
                -- To be nil!
            end
        end

        return frameName
    end

    -- 验证合法性
    local function assertParameters(p)
        TDAssertUtils:assertParameters(p)
    end

    -- 参数赋值
    local function assignParameters(p)
        _resource               = p.resource                            -- 资源名(不包含后缀名的前缀，如要缓存要引入num.plist和num.png，则_resource为num)
        _prefix                 = p.prefix                              -- 数字及符号前缀
        _value                  = p.value                               -- 值
        _resourcePNG            = _resource .. ".png"                   -- 资源图片文件.png
        _resourcePList          = _resource .. ".plist"                 -- 资源信息文件.plist
        _charArray              = TDStringUtils.toCharArray(math.abs(tostring(_value)))       -- 获得字符数组
        _isShowPlus             = p.isShowPlus      or false            -- 显示＋标识
        _isShowMinus            = p.isShowMinus     or false            -- 显示－标识
        _zeroSymbol             = p.zeroSymbol      or M.SYMBOL_NONE    -- 数字0显示符号 可选值：+或-
        _frameNameZeroSymbol    = getFrameName(M.SYMBOLS[_zeroSymbol])  -- 帧名称[0]
        _frameNamePlus          = getFrameName("+")                     -- 帧名称[＋]
        _frameNameMinus         = getFrameName("-")                     -- 帧名称[－]
        _margin                 = p.margin          or M.DFT_MARGIN     -- 项目间隔
        _marginNum              = p.marginNum       or _margin          -- 数字间隔
        _marginSymbol           = p.marginSymbol    or _margin          -- 符号间隔
        _align                  = p.align           or M.ALIGN_CENTER   -- 项目对齐方式 
    end

    -- 加载资源
    local function loadResource()
        display.addSpriteFramesWithFile(_resourcePList, _resourcePNG)
    end

    -- 渲染结点
    local function renderNode(_node)
        if _node then 
            -- [实体渲染]
            if TDArrayUtils.isNotEmpty(_charArray) then
                -- 初始化项目集合
                _items = {}

                -- 符号处理（处理数字前是否加上＋或－）
                if _value == 0 then
                    -- 0值处理 
                    if _frameNameZeroSymbol ~= nil then 
                        _items[#_items + 1] = display.newSprite(_frameNameZeroSymbol)
                    end
                else
                    -- 非0值处理
                    if _value > 0 then 
                        -- 正数处理
                        if _isShowPlus then 
                            _items[#_items + 1] = display.newSprite(_frameNamePlus)
                        end
                    else
                        -- 负数处理
                        if _isShowMinus then 
                            _items[#_items + 1] = display.newSprite(_frameNameMinus)
                        end
                    end
                end
                if _items[#_items] then 
                    -- 符号标记
                    _items[#_items].kind = M.KIND_SYMBOL
                end

                -- 数字处理
                for i = 1, #_charArray do
                    local char              = _charArray[i]
                    local frameName         = getFrameName(tonumber(char))
                    local item              = display.newSprite(frameName)
                    item.kind               = M.KIND_NUM
                    _items[#_items + 1]     = item
                end

                -- 添加并排列位置
                local itemPosition  = ccp(0, 0)
                local itemWidth     = 0
                local totalWidth    = 0
                for i = 1, #_items do
                    local item = _items[i]
                    _node:addChild(item, i, i)
                    item:setPosition(itemPosition)

                    local margin = (item.kind == M.KIND_NUM) and _marginNum or _marginSymbol
                    itemWidth    = item:getContentSize().width * item:getScale()
                    itemPosition = ccp(item:getPositionX() + itemWidth + margin.x, item:getPositionY() + margin.y)
                    totalWidth   = itemPosition.x
                end
                if _align ~= M.ALIGN_NONE then 
                    -- 校正位置X
                    local correctX = _items[1]:getContentSize().width * _items[1]:getScale() * .5
                    if     _align == M.ALIGN_CENTER then
                        correctX  = correctX + totalWidth * -.5
                    elseif _align == M.ALIGN_LEFT   then
                        -- Already right!
                    elseif _align == M.ALIGN_RIGHT  then
                        correctX  = correctX + totalWidth * -1
                    end

                    local offset = ccp(correctX, 0)
                    for i = 1, #_items do
                        local item = _items[i]
                        item:setPosition(ccpAdd(ccp(item:getPosition()), offset))
                    end
                end
            end
        end
    end




    





    -- [具体操作]
    -- [验证参数合法性]
    assertParameters(params)
    -- [参数赋值]
    assignParameters(params)
    -- [加载资源]
    loadResource()

    -- [构建实体]
    -- _node = display.newBatchNode(_resourcePNG, capacity)
    _node = display.newNode()
    if _node then 
        -- -- [实体渲染]
        -- if TDArrayUtils.isNotEmpty(_charArray) then
        --     -- 初始化项目集合
        --     _items = {}

        --     -- 符号处理（处理数字前是否加上＋或－）
        --     if _value == 0 then
        --         -- 0值处理 
        --         if _frameNameZeroSymbol ~= nil then 
        --             _items[#_items + 1] = display.newSprite(_frameNameZeroSymbol)
        --         end
        --     else
        --         -- 非0值处理
        --         if _value > 0 then 
        --             -- 正数处理
        --             if _isShowPlus then 
        --                 _items[#_items + 1] = display.newSprite(_frameNamePlus)
        --             end
        --         else
        --             -- 负数处理
        --             if _isShowMinus then 
        --                 _items[#_items + 1] = display.newSprite(_frameNameMinus)
        --             end
        --         end
        --     end
        --     if _items[#_items] then 
        --         -- 符号标记
        --         _items[#_items].kind = M.KIND_SYMBOL
        --     end

        --     -- 数字处理
        --     for i = 1, #_charArray do
        --         local char              = _charArray[i]
        --         local frameName         = getFrameName(tonumber(char))
        --         local item              = display.newSprite(frameName)
        --         item.kind               = M.KIND_NUM
        --         _items[#_items + 1]     = item
        --     end

        --     -- 添加并排列位置
        --     local itemPosition  = ccp(0, 0)
        --     local itemWidth     = 0
        --     local totalWidth    = 0
        --     for i = 1, #_items do
        --         local item = _items[i]
        --         _node:addChild(item, i, i)
        --         item:setPosition(itemPosition)

        --         local margin = (item.kind == M.KIND_NUM) and _marginNum or _marginSymbol
        --         itemWidth    = item:getContentSize().width * item:getScale()
        --         itemPosition = ccp(item:getPositionX() + itemWidth + margin.x, item:getPositionY() + margin.y)
        --         totalWidth   = itemPosition.x
        --     end
        --     if _align ~= M.ALIGN_NONE then 
        --         -- 校正位置X
        --         local correctX = _items[1]:getContentSize().width * _items[1]:getScale() * .5
        --         if     _align == M.ALIGN_CENTER then
        --             correctX  = correctX + totalWidth * -.5
        --         elseif _align == M.ALIGN_LEFT   then
        --             -- Already right!
        --         elseif _align == M.ALIGN_RIGHT  then
        --             correctX  = correctX + totalWidth * -1
        --         end

        --         local offset = ccp(correctX, 0)
        --         for i = 1, #_items do
        --             local item = _items[i]
        --             item:setPosition(ccpAdd(ccp(item:getPosition()), offset))
        --         end
        --     end
        -- end
        renderNode(_node)

        -- [公有方法定义]
        -- ...
        function _node:update(params)
            -- [验证参数合法性]
            assertParameters(params)
            -- [参数赋值]
            -- assignParameters(params)
            _prefix                 = params.prefix                              -- 数字及符号前缀
            _value                  = params.value                               -- 值
            _align                  = params.align                               -- 排列方式
            _charArray              = TDStringUtils.toCharArray(math.abs(tostring(_value)))       -- 获得字符数组

            -- 清除所有子结点
            self:removeAllChildrenWithCleanup(true)
            -- 渲染结点
            renderNode(self)
        end

        -- ...
        function _node:getValue()
            return _value
        end
    end





    return _node
end



return M