
-- 消息数据库操作

local sqlite3 = require("sqlite3")


local MsgDB = {}

local TAG = "[MsgBD]"


-- 消息数据库存放位置
local MSGDB_PATH = device.documentsPath .. "msgdb.rdb"


--[[ ## 消息数据库模型 --]]

--[[
	## 消息对话表 ##
		该表用于存储所有对话，每一条纪录代表与一位好友的对话。
		并保存是否已读、是否最新、最后一条消息等内容
	Table:  
		talktb 

	Fileds:
		'senderId' 		   		int     		好友id
		'senderNickName'		varchar(100) 	好友昵称
		'isSucceed'			    int				上一条消息是否发送成功(0/1)
		'isnew'				    int				是否最新(0/1)
		'uptime'			    date        	更新时间(用于排序)
		'lastMsgId'			    int				最后一条消息id	
	
	## 消息内容表 ##
		该表用于存储所有对话详细内容，每一条纪录代表一条消息。
	Table:
		msgtb 
	
	Fileds:
		'senderId' 			int 		    发送者id
		'senderNickName' 	varchar(100) 	发送者昵称
		'receiverId'		int  			接收者id
		'receiverName'     varchar(100) 	接收者昵称
		'msgId'				int				消息id
		'vipLevel'			int				发送者等级
		'content'			varchar(500)    消息内容
		'sendTime'			date	     	发送时间
		'messageType'		int				消息类型
		'isSucceed'		    int             是否发送成功
  ]]--
local MSGDB_CREATE_TABLE_TALK_SQL = "create table talktb( "
								.. "id integer primary key autoincrement,"
								.. "senderId integer,"
								.. "senderNickName nvarchar(500),"
								.. "isSucceed integer,"
								.. "isnew integer,"
								.. "uptime date,"
								.. "content TEXT"							
								.. ");"

local MSGDB_CREATE_TABLE_MESSAGE_SQL = "create table msgtb( "
								.. "id integer primary key autoincrement,"
								.. "senderId integer,"
								.. "isBigShow integer,"
								.. "senderNickName nvarchar(500),"
								.. "receiverId integer,"
								.. "receiveId integer,"
								.. "receiverName nvarchar(500),"
								.. "msgId integer,"
								.. "vipLevel integer,"
								.. "messageType integer,"
								.. "sendTime date,"
								.. "content TEXT,"	
								.. "isSucceed integer"						
								.. ");"

-- 消息对话表名
MSGDB_TABLE_TALK = "talktb"

-- 消息内容表名
MSGDB_TABLE_MESSAGE = "msgtb"

-- 消息表
MSGDB_TABLES = {
	
	-- 消息对话表
	TALKTB = {
			TABLE_NAME = MSGDB_TABLE_TALK,

			TABLE_FILED = {
				senderId 	  	  =		"senderId",
				senderNickName	  = 	"senderNickName",
				isSucceed    	  = 	"isSucceed",
				isnew			  =     "isnew",
				uptime    		  =		"uptime",
				content 		  = 	"content",
			},
		},

	-- 消息内容表
	MESSAGETB = {
			TABLE_NAME = MSGDB_TABLE_MESSAGE,

			TABLE_FILED = {
				senderId 	  	  =		"senderId",
				senderNickName    =		"senderNickName",
				receiveId	 	  = 	"receiveId",
				receiverId	 	  = 	"receiverId",
				isBigShow	 	  = 	"isBigShow",
				receiverName	  = 	"receiverName",
				msgId             =     "msgId",
				vipLevel	 	  = 	"vipLevel",
				content           = 	"content",
				sendTime          = 	"sendTime",
				messageType       = 	"messageType",
				isSucceed         = 	"isSucceed",
			},
		}
}


--[[ ## local function --]]

--[[ ## 数据库操作 --]]

-- 数据库 Connection
local _msgdb = nil


--[[  
	判断数据库是否已打开
--]]
local function isOpen( )
	-- body
	return _msgdb ~= NULL
end

--[[ 
	打开消息数据库，数据库默认存放位置为device.documentsPath
	如果打开成功，_db != NULL 
--]]
local function openDB( )
	-- body
	if not isOpen() then
		_msgdb = sqlite3.open(MSGDB_PATH)
	end
end

--[[  
	关闭数据库
--]]
local function closeDB( )
	-- body
	if isOpen() then
		_msgdb:close()
	end
end


--[[ 

在数据库执行SQL语句，并返回执行结果

**Parameters:**

	**require**

-   sql:        string       sql语句
	
	**option**

-   callback:   function     回调
-   userdata:   table        参数

**Returns:**

-   return 执行结果

--]]
local function execSQL( params )
	-- body
	if not isOpen() then
		openDB();
	end

	if not params or not isstring(params.sql) then 
		-- Log.d(TAG , "execSql() invalid param sql")
		return false
	end

	--[[/* stack:
		**  1: db
		**  2: sql
        **  3: callback function
        **  4: userdata
        **  5: column names
        **  6: reusable column values
       */
    --]]
    -- Log.d(TAG , "execSql() sql : " .. params.sql)

	return _msgdb:exec(params.sql)
end


--[[ 

判断数据库种的表是否已存在

**Parameters:**

-   tableName:        string       表名

**Returns:**

-   return true/false

--]]
local function isTableExist( params )
	-- body
	if not isOpen() then
		openDB();
	end

	if not params or not isstring(params.tableName) then 
		-- Log.d(TAG , "isTableExist() invalid param tableName")
		return false
	end

	local _tableName  = params.tableName

	local sqlexc = "SELECT "
					    .. " COUNT(*) as CNT "
				 .. " FROM " 
					    .. " sqlite_master "
			    .. " WHERE "
						.. " type='table' "
						.. " and " 
					    .. " name='" .. _tableName .."'";

	-- print(sqlexc);

	local count
	for result in _msgdb:nrows(sqlexc) 
		do count = result.CNT;
		   -- print(result.CNT);
		   break;
	end
	return count > 0
end


--[[ 
创建表

**Parameters:**

	**require**

- 	tableName:	string 		表名称
-	fileds:		table  		所有字段

	**option**

-   callback:   function     回调
-   userdata:   table        参数

**Returns:**

-	return 执行结果
	
--]]
local function createTable( params )
	-- body
	if not isOpen() then
		openDB();
	end

	if not params or not isstring(params.tableName) then 
		-- Log.d(TAG , "createTable() invalid param tableName")
		return false
	end

	if not istable(params.fileds) then 
		-- Log.d(TAG , "createTable() invalid param fileds")
		return false
	end

	-- 要创建的表名称
	local _tableName = params.tableName
	-- 表字段
	local _fileds 	 = params.fileds

	local _callback  = params.callback
	local _userdata  = params.userdata

	-- 创建表SQL语句
	local sqlexc = "CREATE TABLE ".. _tableName .."( "
	local sqlexcEnd = " )"

	for name,value in pairs(_fileds) do
		-- print(value)
		sqlexc = sqlexc .. value .. " ,"
	end

	if _fileds then
		sqlexc = string.sub(sqlexc,1,#sqlexc - 1) 
	end

	sqlexc = sqlexc .. sqlexcEnd

	-- Log.d(TAG , "createTable() ##sqlexc : " .. sqlexc)

	return execSQL({sql = sqlexc, callback = _callback , userdata = _userdata})
end



-- 是否已数据库连接已打开
function MsgDB.isMessageDBOpen( )
	-- body
	return isOpen();
end


--[[ 

获取表数据

**Parameters:**

	**required**

-   tableName:    string	   表名称
	
	**option**

-   fileds:        table       需要获取的字段, 默认获取全部
-   condition:    string       查询条件

**Returns:**

-   result: 	  table 	   查询结果

--]]
function MsgDB.get( params )
	-- body
	if not isOpen() then
		openDB();
	end

	if not params or not isstring(params.tableName) then 
		-- Log.d(TAG , "get() invalid param tableName")
		return;
	end

	local _tableName = params.tableName
	local _fileds 	 = params.fileds
	local _condition = params.condition

	-- 获取查询字段
	local expresion  = iff( (not _fileds or #_fileds < 1), "*" , table.concat(_fileds, ","))
	
	-- 组合查询语句
	local sqlexc = " SELECT " .. expresion .. 
					 " FROM " .. _tableName ..
					" WHERE 1=1 " .. _condition

	local result = {}

	-- print(sqlexc);

	for value in _msgdb:nrows(sqlexc) do
		-- table.insert (table, [pos,] value)
		table.insert(result, #result+1, value)
	end

	return result
end


--[[ 

删除数据

**Parameters:**

	**require**

-   tableName:    string	   表名称
-   condition:    string       查询条件

	**option**

-   callback:   function     回调
-   userdata:   table        参数

**Returns:**

-   result: true/false

--]]
function MsgDB.delete( params )
	-- body
	if not isOpen() then
		openDB();
	end

	if not params or not isstring(params.tableName) then 
		-- Log.d(TAG , "delete() invalid param tableName")
		return;
	end

	if not params or not isstring(params.condition) then 
		-- Log.d(TAG , "delete() invalid param condition")
		return;
	end

	local _tableName = params.tableName
	local _condition = params.condition

	local sqlexc = "DELETE FROM ".. _tableName .." WHERE 1=1 and ".._condition

	return execSQL({sql = sqlexc, callback = _callback , userdata = _userdata})

end

--[[ 

插入数据

**Parameters:**

	**require**

-   tableName:    string	   表名称
- 	values:		  table 	   要插入的数据
	
	**option**

-   callback:   function     回调
-   userdata:   table        参数

**Returns:**

-   result: true/false

--]]
function MsgDB.insert( params )
	-- body
	if not isOpen() then
		openDB();
	end

	if not params or not isstring(params.tableName) then
		-- Log.a(TAG , "insert() invalid params tableName")
	end

	if not istable(params.values) then
		-- Log.a(TAG , "insert() invalid params values")
	end

	local _tableName = params.tableName  -- 要插入数据的表
	local _values 	 = params.values     -- 要插入的数据

	-- string.format (formatstring, ···)
	local sqlformat = "INSERT INTO %s%s VALUES %s"

	local sqlexc = ""
	local sqlfileds = ""
	local sqlvalues = ""

	local quobegin = "("
	local quoend   = ")"

	-- 组合数据
    for i = 1 , #_values do

    	sqlvalues = quobegin
    	sqlfileds = quobegin

        local tablevalue = _values[i]

        for name,value in pairs(tablevalue) do

   --      	print(name)
			-- print(value)

			sqlfileds = sqlfileds .. name .. " ,"

			if isstring(value) then
				sqlvalues = sqlvalues .."'" .. value .."' ,"
			else
				sqlvalues = sqlvalues .. value ..","
			end
		end

		sqlfileds = string.sub(sqlfileds, 1, #sqlfileds - 1) 
		sqlfileds = sqlfileds .. quoend

		sqlvalues = string.sub(sqlvalues, 1, #sqlvalues - 1) 
		sqlvalues = sqlvalues .. quoend

		sqlexc = sqlexc .. string.format(sqlformat, _tableName, sqlfileds ,sqlvalues) .. ";"

    end

    -- print(sqlexc);

    return execSQL({sql = sqlexc, callback = _callback , userdata = _userdata})

end


--[[ 

更新数据

**Parameters:**

	**require**

-   tableName:    string	   表名称
- 	values:		  array 	   要插入的数据
	
	**option**

-   fileds:       array        缺省字段
-   callback:   function     回调
-   userdata:   table        参数

**Returns:**

-   result: true/false

--]]
function MsgDB.update( params )

	if not isOpen() then
		openDB();
	end

	if not params or not isstring(params.tableName) then 
		-- Log.d(TAG , "update() invalid param tableName")
		return;
	end

	if not params or not isstring(params.condition) then 
		-- Log.d(TAG , "update() invalid param condition")
		return;
	end

	if not params or not isstring(params.update) then 
		-- Log.d(TAG , "update() invalid param update")
		return;
	end

	local _tableName = params.tableName
	local _condition = params.condition
	local _update    = params.update

	local sqlexc = "UPDATE ".. _tableName .." Set ".._update.." WHERE 1=1 and ".._condition

	return execSQL({sql = sqlexc, callback = _callback , userdata = _userdata})

end

--[[ ## function --]]

--[[ 
	初始化
--]]
function MsgDB.init( )
	-- body

	openDB();

	-- 创建对话表
	if not isTableExist({tableName = MSGDB_TABLE_TALK}) then

		execSQL({sql = MSGDB_CREATE_TABLE_TALK_SQL})
	end

	-- 创建消息表
	if not isTableExist({tableName = MSGDB_TABLE_MESSAGE}) then

		execSQL({sql = MSGDB_CREATE_TABLE_MESSAGE_SQL})
	end

	-- -- 关闭数据库

	return true;
end


--[[ 
	finish
--]]
function MsgDB.finish(  )
	-- body

	-- 关闭数据库
	closeDB();

end



--[[ ## Test --]]

--[[ 
	对话数据
--]]
function MsgDB.insertTalks( )
	-- body
	if not isOpen() then
		openDB();
	end

	insert({
			tableName = MSGDB_TABLE_TALK ,
			values 	  = {
							{
								senderId = 1,
								senderNickName = "1",
							},
							{
								senderId = 1,
								senderNickName = "1",
							},
							{
								senderId = 1,
								senderNickName = "1",
							},
							{
								senderId = 1,
								senderNickName = "1",
							},
						}
		})

	-- MsgDB.getAllTalks();

end


return MsgDB