
local anim = {}
local TAG = "[tandyfw.anim]"
--[[ 

播放plist动画

**Parameters:**

-   target:                 CCSprite   矩形
-   filename:               string     plist与png的文件名
-   pattern:                string     格式
-   nums:                   int        总帧数
-   playTime:               int        播放时间(秒)
-   defaultDisplayIndex:    int        默认播放帧
-   forever                 bool       循环播放
-   removeWhenFinished:     bool       播放完成移除(单次播放生效)
-   onComplete:             function   播放完成事件(单次播放生效)
-   delayTime:              int        延迟播放时间(秒)

**Returns:**

-   return target

--]]

function anim.playAnimation( target, filename, pattern, nums, playTime, defaultDisplayIndex, forever, removeWhenFinished, onComplete, delayTime )
    display.addSpriteFramesWithFile( filename .. ".plist", filename .. ".png")
    if not target then
        target = widget.newSprite{file = "#" .. string.format( pattern, defaultDisplayIndex or 1),}
    end
    local frames = display.newFrames(pattern, 1, nums)
    local animation = display.newAnimation(frames, playTime / nums)
    if forever then
        transition.playAnimationForever(target, animation, delayTime)
    else
        transition.playAnimationOnce(target, animation, removeWhenFinished, onComplete, delayTime)
    end 
    return target
end
--[[ 

创建动画,在一套图片中以指定顺序进行动画播放

**Parameters:**

-   filename:       string          引入的png和plist文件名
-   pattern:        string          子图片文件名格式
-   begin:          int             起始下标
-   length:         int             子图片个数
-   time:           float           播放时长
-   ...:            table集          帧序列

**Returns:**

-   return animation

**e.g.**
local animation = anim.createAnimation(
                    "login/login_anim"
                    , "logo_%d.png"
                    , 1, 17
                    , 2.0 / 29,  -- 2.0s播放29帧
                    {1, 9}, {1, 9}, {10, 14}, {14}, {14}, {15, 17}, {10}
                )

--]]
function anim.createAnimation( filename, pattern, begin, length, time, ... )
    display.addSpriteFramesWithFile( filename .. ".plist", filename .. ".png")
    local frames = display.newFrames(pattern, begin, length)
    local function prepareFrames( ... )
        local indexs = { ... }
        local result = {}
        local function prepareFrames_( startIndex, endIndex, step )
            if not endIndex then endIndex = startIndex end
            for i = startIndex, endIndex, step do
                result[#result + 1] = frames[i]
            end
        end
        for _, v in ipairs(indexs) do
            local step = v[3]
            local v2 = v[2] or v[1]
            if not step then 
                if v[1] > v2 then 
                    step = -1 
                else
                    step = 1
                end
            end
            prepareFrames_(v[1], v2, step)
        end
        return result
    end
    local allFrames = nil
    if ... then
        allFrames = prepareFrames(...)
    else
        allFrames = frames
    end
    local animation = display.newAnimation(allFrames, time)
    return animation
end

--[[ 

翻牌效果

**Parameters:**

-   target:                 CCNode     容器
-   playTime:               int        播放总时间(秒)
-   onHalfComplete:         function   播放一半事件(用于完成替换)
-   onComplete:             function   播放完成事件
-   axisX:                  bool       翻转X轴(默认为Y轴)

**Returns:**

-   return target

--]]

function anim.turnPage( target, playTime, onHalfComplete, onComplete, axisX )
    axisX = axisX or false
    local function onHalfComplete_(  )
        onHalfComplete(target)
    end
    transition.execute(target, transition.sequence{
        CCScaleTo:create(playTime / 2, iff(axisX, 1, 0.1), iff(axisX, 0.1, 1)),
        CCCallFunc:create(onHalfComplete_),
        CCScaleTo:create(playTime / 2, 1, 1),
        }, {easing = "IN", onComplete = onComplete })
    return target
end

--[[ 

抖动动画

**Parameters:**

-   target:                 CCNode     目标
-   playTime:               int        播放总时间(秒, 默认0.2s)
-   onComplete:             function   播放完成事件

**Returns:**

-   return target

--]]

function anim.shake( target, args)
    args = args or {}
    local playTime      = (args.playTime or 0.1)/2
    local rotate        = args.rotate or 10.0
    local times         = args.times or 2
    local startDelay    = args.startDelay or 0.0
    local endDelay      = args.endDelay or 0.0
    local forever       = args.forever or false
    local array         = {}
    target:stopAllActions()
    target:setRotation(0.0)
    
    array[#array + 1] = CCDelayTime:create(startDelay)
    for i = 1, times do
        array[#array + 1] = CCRotateTo:create(playTime, rotate)
        array[#array + 1] = CCRotateTo:create(playTime, - rotate)
        array[#array + 1] = CCRotateTo:create(playTime/2 , 0.0)
    end
    array[#array + 1] = CCDelayTime:create(endDelay)
    
    local action = transition.sequence(array)
    if forever then
        action = CCRepeatForever:create(action)
    end
    transition.execute(target, action, {onComplete = args.onComplete})
    return target
end

--[[ 

抖动动画

**Parameters:**

-   target:                 CCNode     目标
-   playTime:               int        播放总时间(秒, 默认0.2s)
-   onComplete:             function   播放完成事件

**Returns:**

-   return target

--]]

function anim.circling( target, args)
    args = args or {}
    local playTime      = args.playTime or 10
    local forward       = args.forward
    local startDelay    = args.startDelay or 0.0
    local endDelay      = args.endDelay or 0.0
    local forever       = args.forever or false

    local rotate = -360
    if forward then
        rotate = 360.0
    end
    
    target:stopAllActions()
    target:setRotation(0.0)

    local array         = {}
    array[#array + 1] = CCRotateBy:create(playTime, rotate)
    
    local action = transition.sequence(array)
    if forever then
        action = CCRepeatForever:create(action)
    end
    transition.execute(target, action, {onComplete = args.onComplete})
    return target
end
--[[ 

淡入淡出

**Parameters:**

-   target:                 CCNode     目标
-   playTime:               int        播放总时间(秒, 默认0.2s)
-   onComplete:             function   播放完成事件

**Returns:**

-   return target

--]]

function anim.fadeFade( target, args)
    args = args or {}
    local playTime      = args.playTime or 1
    local max           = args.max or 255
    local min           = args.min or 0
    local startDelay    = args.startDelay or 0.0
    local endDelay      = args.endDelay or 0.0
    local forever       = args.forever or false
    
    -- target:stopAllActions()

    local array         = {}
    array[#array + 1]   = CCDelayTime:create(startDelay)
    array[#array + 1]   = CCFadeTo:create(playTime, max)
    array[#array + 1]   = CCFadeTo:create(playTime, min)
    array[#array + 1]   = CCDelayTime:create(endDelay)

    local action = transition.sequence(array)
    if forever then
        action = CCRepeatForever:create(action)
    end
    transition.execute(target, action, {onComplete = args.onComplete})
    return target
end

--[[ 

Pop动画

**Parameters:**

-   target:                 CCNode     目标
-   playTime:               int        播放总时间(秒, 默认0.2s)
-   onComplete:             function   播放完成事件

**Returns:**

-   return target

--]]

function anim.popOut( target, args)
    args = args or {}
    local playTime = args.playTime or 0.2
    target:setScale(0)
    transition.execute(target, transition.sequence{
        CCScaleTo:create(playTime , 1, 1),
        }, {easing = "CCEaseBackOut", onComplete = args.onComplete })
    return target
end

--[[ 

从底部移出

**Parameters:**

-   target:                 CCNode     目标
-   playTime:               int        播放总时间(秒, 默认0.2s)
-   onComplete:             function   播放完成事件

**Returns:**

-   return target

--]]

function anim.bottomOut( target, args )
    args = args or {}
    local playTime = args.playTime or 0.2
    transition.execute(target, transition.sequence{
        CCMoveTo:create(playTime, ccp(0, -1 * target:getContentSize().height)),
        }, { onComplete = args.onComplete })
    return target
end

--[[ 

从底部进入

**Parameters:**

-   target:                 CCNode     目标
-   playTime:               int        播放总时间(秒, 默认0.2s)
-   onComplete:             function   播放完成事件

**Returns:**

-   return target

--]]

function anim.bottomIn( target, args)
    args = args or {}
    local playTime = args.playTime or 0.2
    target:setPosition(ccp(0, -1 * target:getContentSize().height))
    transition.execute(target, transition.sequence{
        CCMoveTo:create(0.2,ccp(0,0)),
        }, { onComplete = args.onComplete })
    return target
end

--[[ 

从顶部退出

**Parameters:**

-   target:                 CCNode     目标
-   playTime:               int        播放总时间(秒, 默认0.2s)
-   onComplete:             function   播放完成事件

**Returns:**

-   return target

--]]

function anim.topOut( target, args)
    args = args or {}
    local playTime = args.playTime or 0.2
    transition.execute(target, transition.sequence{
        CCMoveTo:create(playTime, ccp(0, target:getContentSize().height)),
        }, { onComplete = args.onComplete })
    return target
end

--[[ 

从顶部进入

**Parameters:**

-   target:                 CCNode     目标
-   playTime:               int        播放总时间(秒, 默认0.2s)
-   onComplete:             function   播放完成事件

**Returns:**

-   return target

--]]

function anim.topIn( target, args)
    args = args or {}
    args.startPoi = args.startPoi or ccp(0, target:getContentSize().height)
    args.endPoi = args.endPoi or ccp(0, 0)
    local playTime = args.playTime or 0.2
    target:setPosition(args.startPoi)
    transition.execute(target, transition.sequence{
        CCMoveTo:create(0.2, args.endPoi),
        }, { onComplete = args.onComplete })
    return target
end

--[[ 

从底部渐隐入

**Parameters:**

-   target:                 CCNode     目标
-   playTime:               int        播放总时间(秒, 默认0.2s)
-   onComplete:             function   播放完成事件

**Returns:**

-   return target

--]]

function anim.bottomFadeIn( target, args)
    args = args or {}
    local pos = args.pos or 20
    local playTime = args.playTime or 0.2
    local position = ccp(target:getPositionX(), target:getPositionY())
    target:setVisible(true)
    target:setOpacity(0)
    target:setPosition(ccp(position.x, position.y - pos))
    transition.execute(target, transition.sequence{
        CCSpawn:createWithTwoActions(CCMoveTo:create(playTime, position), CCFadeIn:create(playTime)),
        }, { onComplete = args.onComplete })
    return target
end

--[[ 

从底部渐隐出

**Parameters:**

-   target:                 CCNode     目标
-   playTime:               int        播放总时间(秒, 默认0.2s)
-   onComplete:             function   播放完成事件

**Returns:**

-   return target

--]]

function anim.bottomFadeOut( target, playTime, onComplete, pos)
    args = args or {}
    local pos = args.pos or 20
    local playTime = args.playTime or 0.2
    local position = ccp(target:getPositionX(), target:getPositionY())
    target:setOpacity(255)
    transition.execute(target, transition.sequence{
        CCSpawn:createWithTwoActions(CCMoveTo:create(playTime, ccp(position.x, position.y - pos)), CCFadeOut:create(playTime)),
        }, { onComplete = function ( )
            target:setPosition(position)
            if args.onComplete then args.onComplete() end
        end })
    return target
end

--[[动画设定]]--
anim.ANIMATION_CUSTOM = "custom"
anim.ANIMATION_POP_OUT = "popOut"
anim.ANIMATION_TOP_IN = "topIn"
anim.ANIMATION_TOP_OUT = "topOut"
anim.ANIMATION_BOTTOM_IN = "bottomIn"
anim.ANIMATION_BOTTOM_OUT = "bottomOut"
anim.ANIMATION_BOTTOM_FADE_IN = "bottomFadeIn"
anim.ANIMATION_BOTTOM_FADE_OUT = "bottomFadeOut"

function anim.playNodeAnimation( target, animType, args )
    if target and animType then 
        args = args or {}
        if anim[animType] then 
            anim[animType](target, args)
        elseif animType == anim.ANIMATION_CUSTOM then
            if type(args.fn) ~= "function" then 
                args.fn(target, args)
            end 
        end
    end
end























return anim