--[[ 

Entity/POJO 基类,对ModelBase进行修改,支持直接以table参数形式对类中类进行定义,包括单个的和列表的
定义包含BaseEntity子类对象的方式
    xxxEntity = { "xxx.xxx.xxxEntity" } ,
定义包含BaseEntity子类对象列表的方式,1说明其为列表,3指明其类型(只能一种),最后两个是可选项,用于指定需要导出的字段及过滤条件
    xxxEntitys = { "table", nil, "xxx.xxx.xxxEntity",  fields = {...}, filter = {...} }
--]]

local json = require("framework.shared.json")

local BaseEntity = class("BaseEntity")
BaseEntity.idkey = "id"
BaseEntity.schema = {
    id = {"string"},
    --定义包含BaseEntity子类对象的方式
    --xxxEntity = { "xxx.xxx.xxxEntity" } ,
    --定义包含BaseEntity子类对象列表的方式,1说明其为列表,3指明其类型(只能一种),最后两个是可选项,用于指定需要导出的字段及过滤条件
    --xxxEntitys = { "table", nil, "xxx.xxx.xxxEntity",  fields = {...}, filter = {...} }
}
BaseEntity.fields = {"id"}

local function filterProperties(properties, filter)
    for i, field in ipairs(filter) do
        properties[field] = nil
    end
end

-- local function upperFirst( str )
--     if not str then return str end
--     return string.upper(string.sub(str, 1, 1)) .. string.sub(str,2)
-- end
--[[ 

创建属性的get set集

--]]
-- local function createPropertyAccessPair(self)
--     for field, schema in pairs(self.class.schema) do
--         local typ, def = unpack(schema)
--         local propname = field .. "_"
--         local upperPropname = string.ucfirst(field)
--         if typ == "boolean" then
--             self["is" .. upperPropname] = function (self)
--                 return self[propname]
--             end
--             self["set" .. upperPropname] = function (self,a)
--                 self[propname] = a
--             end
--         else
--             self["get" .. upperPropname] = function (self)
--                 return self[propname]
--             end
--             self["set" .. upperPropname] = function (self, a)
--                 self[propname] = a
--             end
--         end
--     end
-- end

function BaseEntity:ctor(properties)
    self.isBaseEntity_ = true
    if type(properties) ~= "table" then properties = {} end
    self:setProperties(properties)
end

--[[--

返回对象的 ID 值。

**Returns:**

-   ID 值

]]
function BaseEntity:getId()
    local id = self[self.class.idkey .. "_"]
    assert(id ~= nil, string.format("%s [%s:getId()] Invalid id", tostring(self), self.class.classname))
    return id
end

--[[--

确定对象是否设置了有效的 Id。

]]
function BaseEntity:isValidId()
    local propname = self.class.idkey .. "_"
    local id = self[propname]
    return type(id) == "string" and id ~= ""
end

--[[--

修改对象的属性。

**Parameters:**

-   properties: 包含属性值的数组

]]
function BaseEntity:setProperties(properties)
    assert(type(properties) == "table", "Invalid properties")
           -- string.format("%s [%s:setProperties()] Invalid properties", tostring(self), self.class.classname))

    -- if not self.hasCreatePropertyAccessPair_ then 
    --     createPropertyAccessPair(self)
    --     self.hasCreatePropertyAccessPair_ = true
    -- end

    for field, schema in pairs(self.class.schema) do
        local typ, def = unpack(schema)
        local propname = field .. "_"

        local val = properties[field] or def
        if val ~= nil then
            if typ == "number" then val = tonumber(val) end
            --[=[
            当类型为table时有三种情况可以处理
            1.定义的不是table,而是BaseEntity的子类对象的require路径,用于one to one
            2.定义的是table,并在schema中定义了第三个表示BaseEntity的子类对象的require路径,用于one to many
            3.普通定义,就是个table,同原样
            --]=]
            if type(val) == "table" and typ ~= "table" then
                --针对只表示一个BaseEntity子类对象
                self[propname] = require(typ).new(val)
            elseif type(val) == "table" and typ == "table" and schema[3] then
                --针对BaseEntity子类对象
                local vals = {}
                local EntityClass = require(schema[3])
                for i = 1, #val do
                    local entity = EntityClass.new(val[i])
                    table.insert(vals, entity)
                end
                self[propname] = vals
            else
                assert(type(val) == typ,
                   string.format("%s [%s:setProperties()] Type mismatch, %s expected %s, actual is %s",
                                 tostring(self), self.class.classname, field, typ, type(val)))
                self[propname] = val
            end

            
        end
    end
end

--[[--

取得对象的属性值。

**Parameters:**

-   fields: 要取得哪些属性的值，如果未指定该参数，则返回 fields 中设定的属性
-   filter: 要从结果中过滤掉哪些属性，如果未指定则不过滤

**Returns:**

-   包含属性值的数组

]]
function BaseEntity:getProperties(fields, filter)
    local schema = self.class.schema
    if type(fields) ~= "table" then fields = self.class.fields end

    local properties = {}
    for i, field in ipairs(fields) do
        local propname = field .. "_"
        local typ = schema[field][1]
        local val = self[propname]
        if type(val) == "table" and typ ~= "table" then
            properties[field] = val:getProperties(schema[field].fields, schema[field].filter)
        elseif type(val) == "table" and typ == "table" and schema[field][3] then
            local vals = {}
            for i = 1, #val do
                table.insert(vals, val[i]:getProperties(schema[field].fields, schema[field].filter))
            end
            properties[field] = vals
        else
            if val ~= nil then
                assert(type(val) == typ,
                   string.format("%s [%s:getProperties()] Type mismatch, %s expected %s, actual is %s",
                                     tostring(self), self.class.classname, field, typ, type(val)))
                properties[field] = val
            end
        end
    end

    if type(filter) == "table" then
        filterProperties(properties, filter)
    end

    return properties
end

--[[--

生成JSON

**Parameters:**

-   fields: 要取得哪些属性的值，如果未指定该参数，则返回 fields 中设定的属性
-   filter: 要从结果中过滤掉哪些属性，如果未指定则不过滤

**Returns:**

-   包含属性值的数组

]]
function BaseEntity:toJSON(fields, filter)
    return json.encode(self:getProperties(fields, filter))
end

return BaseEntity
