
local util = {}
local TAG = "[tandyfw.util]"

--[[ 

设置显示尺寸

**Parameters:**

-   width:          int      宽度
-   height:         int      高度
-   mode:           enum     显示模式
                            kResolutionExactFit
                            kResolutionNoBorder
                            kResolutionShowAll (默认)

--]]
function util.setDesignResolutionSize( width, height, mode )
    CCDirector:sharedDirector():getOpenGLView():setDesignResolutionSize(width, height, mode or kResolutionShowAll)
end
display.setDesignResolutionSize = util.setDesignResolutionSize
--[[ 

判断点是否在矩形内

**Parameters:**

-   rect:        CCRect     矩形
-   point:       CCPoint    点

**Returns:**

-   return 点是否在矩形内

--]]
function util.rectContainsPoint( rect, point )
    return (point.x >= rect:getMinX() and point.x <= rect:getMaxX()
        and point.y >= rect:getMinY() and point.y <= rect:getMaxY())
end
--[[ 

判断point是否在四边形内GL坐标系上
算法:根据向量S = (x1-x3)*(y2-y3)-(y1-y3)(x2-x3),
判断点(x3, y3)与由点(x1, y1),(x2, y2)连成的线的位置,
S大于0在左边,小于0在右边,等于0在线上

**Parameters:**

-   left:       CCPoint     左顶点
-   top:        CCPoint     上顶点
-   right:      CCPoint     右顶点
-   bottom:     CCPoint     下顶点
-   point:      CCPoint     要判断的点
        
**Returns:**
        
-   return bool 结果
        
--]]
function util.quadContainsPointGL( left, top, right, bottom, point )
    local value = 1
    -- 第一象限
    if point.x >= top.x and point.y <= right.y then
        value = (right.x - point.x)*(top.y - point.y) - (right.y - point.y)*(top.x - point.x)
    -- 第二象限
    elseif point.x < top.x and point.y < left.y then
        value = (top.x - point.x)*(left.y - point.y)-(top.y - point.y)*(left.x - point.x)
    -- 第三象限
    elseif point.x <= bottom.x and point.y >= left.y then
        value = (left.x - point.x)*(bottom.y - point.y)-(left.y - point.y)*(bottom.x - point.x) 
    -- 第四象限 
    elseif point.x > bottom.x and point.y > right.y then
        value = (bottom.x - point.x)*(right.y - point.y)-(bottom.y - point.y)*(right.x - point.x)
    end
    if value <= 0 then
        return true
    end
    return false
end    
--[[ 

获取node中点

**Parameters:**

-   node:        CCNode     容器

**Returns:**

-   中心点坐标

--]]
function util.centerNode( node )
	if node then
		return node:getContentSize().width / 2, node:getContentSize().height / 2
	end
	return 0, 0
end
--[[ 

图片加入缓存

**Parameters:**

-   fileName:        string     文件名

**Returns:**

-   缓存自身返回

--]]
function util.addTextureCache( fileName )
	return CCTextureCache:sharedTextureCache():addImage(fileName)
end

function util.copyNodeProperties( srcNode, destNode )
    Log.a(isnotnil(srcNode) and isnotnil(destNode), TAG, "copyNodeProperties() missing srcNode and/or destNode")
    --contentsize
    destNode:setContentSize(srcNode:getContentSize())
    --position
    destNode:setPosition(srcNode:getPosition())
    --anchor
    destNode:setAnchorPoint(srcNode:getAnchorPoint())
    --ignor
    destNode:ignoreAnchorPointForPosition(srcNode:isIgnoreAnchorPointForPosition())
    --tag
    destNode:setTag(srcNode:getTag())
end

--[[ 

延迟调用指定方法

**Parameters:**

-   fn:        function     将调用的方法
-   time:       number      延迟时间,默认为0.5s 

--]]
function util.delayCall( fn, time )
    if time == 0 then 
        fn()
        return nil
    else
        return scheduler.performWithDelayGlobal(fn, time or 0.5)
    end
end
--[[ 

非立即调用方法,用于函数式编程时传递函数

**Parameters:**


-   fn:        function     将调用的方法
-   ...:        ...         附带的参数

**Returns:**

-   return function

--]]
function util.callFun( fn, ... )
    local args = {...}
    return function() fn(unpack(args)) end
end
--[[ 

延迟调用指定方法,带参数

**Parameters:**

-   time:       number      延迟时间,默认为0.5s 
-   fn:        function     将调用的方法
-   ...:        ...         附带的参数

e.g.
local function printInfo( ... ) print(...) end
util.delayCallWithArgs(10, printInfo, 5, "a", 7)

--]]
function util.delayCallWithArgs( time, fn, ... )
    return scheduler.performWithDelayGlobal(callFun(fn, ...), time or 0.5)
end
--[[ 

交替显示Layer,用于两Layer在同一位置,一次只显示其中一个而隐藏另一个的情况

**Parameters:**

-   willShowLayer:        CCNode     将要显示的Layer
-   willHideLayer:        CCNode     将要隐藏的Layer
-   swap:                 boolean    前两个参数互换,默认为false

--]]
function util.changeDisplayLayer( willShowLayer, willHideLayer, swap )
    local showLayer = iff(swap, willHideLayer, willShowLayer)
    local hideLayer = iff(swap, willShowLayer, willHideLayer)
    if hideLayer:isVisible() then
        local position = hideLayer:getPositionLua()
        hideLayer:setPosition(0, 10000)
        hideLayer:setVisible(false)
        showLayer:setPosition(position)
        showLayer:setVisible(true)
    end
end
--[[ 

同一容器中显示出其中一个Layer,其它的移出可视区域并隐藏

**Parameters:**

-   container:          CCNode          容器
-   displayNode:        CCNode/int          将要显示的Node

--]]
function util.showSingleLayer( container, displayNode )
    if container and displayNode then
        if isnumber(displayNode) then
            displayNode = container:getChildByTag(displayNode)
        else
            Log.a(displayNode:getParent() == container, TAG, "showSingleLayer() Wrong displayNode")
        end
        local array = container:getChildren()
        local position 
        for i = 0, container:getChildrenCount() - 1 do
            local layer = array:objectAtIndex(i)
            layer = util.toCCNode(layer)
            if layer then
                --找到所有Layer,移出显示区域并隐藏
                if layer:isVisible() then
                    --小登记下位置
                    position = layer:getPositionLua()
                    layer:setVisible(false)
                end
                layer:setPosition(0, 10000)
            end
        end
        --将要显示的移到显示区域内并显示出来
        if position then
            displayNode:setVisible(true)
            displayNode:setPosition(position)
        end
    end
end
--[[ 

显示/隐藏指定Layer(同时隐藏其包含的CCControlButton,防止其点击事件触发)

**Parameters:**

-   layer:          CCNode              指定要操作的Layer
-   display:        bool                是否显示,默认为true

--]]
function util.displayLayer( layer, display )
    if layer then
        display = ifnil(display, true)
        if layer:getChildrenCount() > 0 then
            local array = layer:getChildren()
            for i = 0, layer:getChildrenCount() - 1 do
                local obj = array:objectAtIndex(i)
                obj = util.toCCControlButton(obj)
                if obj then 
                    if display then
                        if obj.LUA_CCCONTROLBUTTON then
                            obj.LUA_CCCONTROLBUTTON = nil
                            obj:setVisible(display)
                        end
                    else
                        obj.LUA_CCCONTROLBUTTON = true
                        obj:setVisible(display)
                    end
                end
            end
        end
        layer:setVisible(display)
    end
end
--[[ 

开启点击事件/关闭点击事件指定Layer(同时关闭其包含的CCControlButton,防止其点击事件触发)

**Parameters:**

-   layer:          CCNode              指定要操作的Layer
-   enabled:        bool                是否Enabled,默认为true

--]]
function util.enabledLayer( layer, enabled )
    if layer then
        enabled = ifnil(enabled, true)
        if layer:getChildrenCount() > 0 then
            local array = layer:getChildren()
            for i = 0, layer:getChildrenCount() - 1 do
                local obj = array:objectAtIndex(i)
                obj = util.toCCControlButton(obj)
                if obj then 
                    if enabled then
                        if obj.LUA_CCCONTROLBUTTON then
                            obj.LUA_CCCONTROLBUTTON = nil
                            obj:setEnabled(enabled)
                        end
                    else
                        obj.LUA_CCCONTROLBUTTON = true
                        obj:setEnabled(enabled)
                    end
                end
            end
        end
    end
end
--[[ 

判断指定node的N级父容器是否为可见状态

**Parameters:**

-   node:        CCNode     用于判定的node

**Returns:**

-   return N级父容器是否为可见状态

--]]
function util.isNParentVisible( node )
    Log.a(node, TAG, "isNParentVisible() invalid node")
    local result = true
    while true do
        if node then
            result = node:isVisible()
            if not result then
                return false
            end
            node = node:getParent()
        else
            break
        end
    end
    return result
end
--[[ 

判断指定node的N级父容器是否为可见并可点击状态

**Parameters:**

-   node:        CCNode     用于判定的node

**Returns:**

-   return N级父容器是否为可见并可点击状态

--]]
function util.isCustomTouchEnabled( node )
    Log.a(node, TAG, "isCustomTouchEnabled() invalid node")
    local result = true
    while true do
        if node then
            result = node:isVisible()
            if result and node._customTouchEnabled == false then
                result = false
            end
            if not result then
                return false
            end
            node = node:getParent()
        else
            break
        end
    end
    return result
end

--[[ 

设定指定node是否可自定义touch

**Parameters:**

-   node:        CCNode     设定的node
-   enable:      boolean    是否启用,默认为true

--]]
function util.enableCustomTouch( node, enable )
    node._customTouchEnabled = iff(enable == nil, true, enable)
end
--[[ 

布局node,将node放置到指定size的9个关键点上,会同时修改锚点

**Parameters:**

-   node:           CCNode          待放置的node
-   size:           CCSize|CCNode   其放置容器的尺寸,当然你放个有尺寸的node也行
-   anchor:         enum            锚点,参看display,默认为display.CENTER
-   offset:         int|table       偏移,默认为0,以下是针对x,y点做偏移的方式:(如果这个是table,则自动调用到util.alignNodeWithPadding)
                                        display.TOP_LEFT:       +offset, -offset
                                        display.CENTER_LEFT:    +offset, 0
                                        display.BOTTOM_LEFT:    +offset, +offset 
                                        display.TOP_CENTER:     0, -offset
                                        display.CENTER:         0, 0
                                        display.BOTTOM_CENTER:  0, offset
                                        display.TOP_RIGHT:      -offset, -offset
                                        display.CENTER_RIGHT:   -offset, 0
                                        display.BOTTOM_RIGHT:   -offset, +offset

--]]
function util.alignNode( node, size, anchor, offset )
    Log.a(node, TAG, "alignNode() invalid params: node")
    Log.a(size, TAG, "alignNode() invalid params: size")
    if istable(offset) then 
        -- 这样可以统一入口
        util.alignNodeWithPadding( node, size, anchor, offset)
        return 
    end
    local x, y
    if size.getContentSize then 
        size = size:getContentSize()
    end
    local w, h = size.width, size.height
    anchor = anchor or display.CENTER
    offset = offset or 0
    if anchor == display.TOP_LEFT or anchor == display.LEFT_TOP then
        x, y = offset, h - offset
    elseif anchor == display.CENTER_LEFT or anchor == display.LEFT_CENTER then
        x, y = offset, h / 2
    elseif anchor == display.BOTTOM_LEFT or anchor == display.LEFT_BOTTOM then
        x, y = offset, offset
    elseif anchor == display.CENTER then
        x, y = w / 2, h / 2
    elseif anchor == display.BOTTOM_RIGHT or anchor == display.RIGHT_BOTTOM then
        x, y = w - offset, offset
    elseif anchor == display.CENTER_RIGHT or anchor == display.RIGHT_CENTER then
        x, y = w - offset, h / 2
    elseif anchor == display.TOP_CENTER or anchor == display.CENTER_TOP then
        x, y = w / 2, h - offset
    elseif anchor == display.BOTTOM_CENTER or anchor == display.CENTER_BOTTOM then
        x, y = w / 2, offset
    else
        x, y = w - offset, h - offset
    end
    node:ignoreAnchorPointForPosition(false)
    display.align(node, anchor, x, y)
end
--[[ 

布局node,将node放置到指定size的9个关键点上,会同时修改锚点

**Parameters:**

-   node:           CCNode          待放置的node
-   size:           CCSize|CCNode   其放置容器的尺寸,当然你放个有尺寸的node也行
-   anchor:         enum            锚点,参看display,默认为display.CENTER
-   padding:        table           四方向间隔,默认都为0, 以下是针对x,y点做偏移的方式
                                        display.TOP_LEFT:       +left, -top
                                        display.CENTER_LEFT:    +left, -top
                                        display.BOTTOM_LEFT:    +left, +bottom 
                                        display.TOP_CENTER:     -left, -top
                                        display.CENTER:         +left, -top
                                        display.BOTTOM_CENTER:  -left, +bottom
                                        display.TOP_RIGHT:      -right, -top
                                        display.CENTER_RIGHT:   -right, -top
                                        display.BOTTOM_RIGHT:   -right, +bottom

--]]
function util.alignNodeWithPadding( node, size, anchor, padding )
    Log.a(node, TAG, "alignNodeWithPadding() invalid params: node")
    Log.a(size, TAG, "alignNodeWithPadding() invalid params: size")
    local x, y
    if size.getContentSize then 
        size = size:getContentSize()
    end
    local w, h = size.width, size.height
    anchor = anchor or display.CENTER
    padding = padding or {}
    padding.left = padding.left or 0
    padding.top = padding.top or 0
    padding.right = padding.right or 0
    padding.bottom = padding.bottom or 0

    if anchor == display.TOP_LEFT or anchor == display.LEFT_TOP then
        x, y = padding.left, h - padding.top
    elseif anchor == display.CENTER_LEFT or anchor == display.LEFT_CENTER then
        x, y = padding.left, h / 2 - padding.top
    elseif anchor == display.BOTTOM_LEFT or anchor == display.LEFT_BOTTOM then
        x, y = padding.left, padding.bottom
    elseif anchor == display.CENTER then
        x, y = w / 2 + padding.left, h / 2 - padding.top
    elseif anchor == display.BOTTOM_RIGHT or anchor == display.RIGHT_BOTTOM then
        x, y = w - padding.right, padding.bottom
    elseif anchor == display.CENTER_RIGHT or anchor == display.RIGHT_CENTER then
        x, y = w - padding.right, h / 2 - padding.top
    elseif anchor == display.TOP_CENTER or anchor == display.CENTER_TOP then
        x, y = w / 2 - padding.left, h - padding.top
    elseif anchor == display.BOTTOM_CENTER or anchor == display.CENTER_BOTTOM then
        x, y = w / 2 - padding.left, padding.bottom
    else
        x, y = w - padding.right, h - padding.top
    end
    node:ignoreAnchorPointForPosition(false)
    display.align(node, anchor, x, y)
end

--[[ 

判断指定坐标是否在指定node里

**Parameters:**

-   node:           CCNode      待判断的node
-   x:              int         world坐标中的x
-   y:              int         world坐标中的y

**Returns:**    

-   return 指定坐标是否在指定node中

--]]
function util.isPointInNode( node, x, y )
    local touchPoint 
    if isnumber(x) then
        touchPoint = ccp(x, y)
    else
        touchPoint = x
    end
    local point = node:convertToNodeSpace(touchPoint)
    local rect = node:boundingBox()
    if node:getScaleX() == node:getScaleY() and node:getScale() ~= 1 then
        local srcScale = 1 / node:getScale() 
        rect.size = CCSizeMul(rect.size, srcScale, srcScale)
    end
    rect.origin = CCPointZero
    return rect:containsPoint(point)
end
--[[ 

自动调整ScrollView,将内容的左上角定位在ScrollView的左上角

**Parameters:**

-   scrollView:        CCScrollView      待调整的ScrollView
-   bounceable:        bool              是否启用边界阻尼效果,默认true

**Returns:**

-   return 是否成功

--]]
function util.autoFixScrollView( scrollView, bounceable )
    local sv = util.toCCScrollView(scrollView)
    if sv then
        sv:setContentOffset(ccp(
            0, 
            sv:getViewSize().height - sv:getContainer():getContentSize().height * sv:getContainer():getScale()
            ))
        if sv:isBounceable() ~= ifnil(bounceable, true) then 
            sv:setBounceable(bounceable) 
        end
    end
    return sv
end
--[[ 

遍历启用/停用容器内控件

**Parameters:**

-   container:        CCNode      待遍历容器
-   enable:           bool        是否启用,默认为true

--]]
function util.enableWidgetInContainer( container, enable )
    -- local containers = { container }
    enable = ifnil(enable, true)
    -- print("===========================================", enable)
    local function enableWidgetInContainer_( subContainer )
        if subContainer:getChildrenCount() > 0 then 
            local children = subContainer:getChildren()
            for i = 0, subContainer:getChildrenCount() - 1 do
                local node = children:objectAtIndex(i)
                local nodeType = tolua.type(node)
                if nodeType == "CCObject" then 
                    node = tolua.cast(node, "CCNode")
                end
                -- print("............node.NODE_TYPE", node.NODE_TYPE)
                if node.NODE_TYPE == "CCControlButton" or node.NODE_TYPE == "CCEditBox" or nodeType == "CCControlButton" or nodeType == "CCEditBox" then
                    local btn = tolua.cast(node, "CCControlButton")
                    if enable == false and btn:isEnabled() then 
                        btn.ENABLED_CUSTOM = true
                        btn:setEnabled(enable)
                    elseif enable == true and btn.ENABLED_CUSTOM == true then
                        btn.ENABLED_CUSTOM = nil
                        btn:setEnabled(enable)
                    end
                -- elseif nodeType == "CCEditBox" then
                --     local editBox = tolua.cast(node, "CCEditBox")
                --     editBox:setVisible(enable)
                elseif nodeType == "CCObject" or nodeType == "CCNode" or nodeType == "CCLayer" then
                    enableWidgetInContainer_(node)
                end
            end
        end
        -- local node = table.remove(containers)
        -- if node then 
        --     if node:getChildrenCount() > 0 then
        --         local children = node:getChildren()
        --         for i = 0, node:getChildrenCount() - 1 do
        --             local child = children:objectAtIndex(i)
        --             local nodeType = tolua.type(child)
        --             print("type:", nodeType)
        --             if nodeType == "CCObject" then 
        --                 child = tolua.cast(child, "CCNode")
        --             end
        --             if nodeType == "CCControlButton" then
        --                 local btn = tolua.cast(child, "CCControlButton")
        --                 btn:setVisible(false)
        --             elseif nodeType == "CCObject" or nodeType == "CCNode" or nodeType == "CCLayer" then
        --                 table.insert(containers, child)
        --             end 
        --         end
        --     end
        -- end
        -- if #containers > 0 then enableWidgetInContainer_() end
    end
    enableWidgetInContainer_(container)

    -- print("===========================================")
end

function util.startDebug( ip )
    if DEVELOP_MOBDEBUG == true then
        require("mobdebug").start( ip or "127.0.0.1" )
    end
end
--[[ 

复制文件

**Parameters:**

-   src:        string       源文件相对路径
-   dest:       string       目的文件路径
-   overwrite:  boolean      是否允许覆盖,默认为true

**Returns:**

-   return 是否成功

--]]
function util.copyfile( src, dest, overwrite )
    src = CCFileUtils:sharedFileUtils():fullPathForFilename(src)
    -- dest = (PTAH_STORAGE_RES or device.documentsPath) .. dest
    Log.a(io.exists(src), TAG, "copyFile() src file %s is not exist.", src)
    if io.exists(dest) then 
        Log.w(TAG, "copyFile() dest file %s is exist.", dest)
        if not ifnil(overwrite, true) then return false end
    end
    local inputFile = io.open(src, "rb")
    if inputFile then 
        local content = inputFile:read("*all")
        io.close(inputFile)
        io.mkdir(dest)
        local outputFile = io.open(dest, "wb")
        if outputFile then 
            local result = outputFile:write(content) ~= nil
            io.close(outputFile)
            return result
        end
    end
    return false
end
io.copyfile = util.copyfile

--[[ 

创建文件夹

**Parameters:**

-   dirname:        string      文件或文件夹的路径

**Returns:**

-   return 成功与否

--]]
function io.mkdir(dirname)
    dirname = util.fullPath(dirname)
    if dirname then 
        local last1 = string.findlast(dirname, "/")
        local last2 = string.findlast(dirname, "%.")
        if last1 and last2 then
            if last1 < last2 then 
                -- 最后的/后有点,把这个当作是文件好了
                dirname, _ = string.splitlast(dirname, "/")
            end
        end
        local lfs = require("lfs")
        return lfs.mkdir(dirname)
    end
    return false
end

--[[ 

删除文件或文件夹, 不存在则无反应

**Parameters:**

-   fullPath:        string      文件或文件夹的绝对路径

--]]
io.remove = function ( fullPath )
    os.execute("rm -r -f \"" .. fullPath .. "\"")
end

--[[ 

在android时调整位置

**Parameters:**
    
-   node :      CCNode          需要调整的node
-   dx :        float           x方向调整值
-   dy :        float           y方向调整值

--]]
function util.modifyPositionInAndroid( node, dx, dy )
    if isandroid() and node and (isnumber(dx) or isnumber(dy)) then 
        dx = dx or 0
        dy = dy or 0
        if dx == 0 and dy == 0 then return end
        node:setPosition(node:getPositionX() + dx, node:getPositionY() + dy)
    end
end

--[[ 

移除未使用的texture缓存

--]]
function util.removeUnusedTextures()
    CCSpriteFrameCache:sharedSpriteFrameCache():removeUnusedSpriteFrames()
    CCTextureCache:sharedTextureCache():removeUnusedTextures()
    -- Log.d("util.removeUnusedTextures", "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    -- Log.d("util.removeUnusedTextures", "@@@ removeUnusedTextures ")
    -- Log.d("util.removeUnusedTextures", "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    -- CCTextureCache:sharedTextureCache():dumpCachedTextureInfo()
    -- Log.d("util.removeUnusedTextures", "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    -- Log.d("util.removeUnusedTextures", "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
end

function display.addSpriteFrame( fileName )
    local frameName = fileName
    if string.find(fileName, "/") then 
        _, frameName = string.splitlast(fileName, "/")
    end
    local cache = CCSpriteFrameCache:sharedSpriteFrameCache()
    local frame = cache:spriteFrameByName(frameName)
    if not frame then
        local texture = CCTextureCache:sharedTextureCache():addImage(fileName)
        if texture then 
            local size = texture:getContentSizeInPixels()
            frame = CCSpriteFrame:createWithTexture(texture, CCRectMake(0, 0, size.width, size.height))
            cache:addSpriteFrame(frame, frameName)
        end
    end
    return frame
end

--[[ 

解压zip文件

**Parameters:**

-   zipFilePath:        string       zip文件绝对路径
-   destPath:           string       解压路径

**Returns:**

-   return 是否成功

--]]
function util.unzip( zipFilePath, destPath )
    return FileManager:uncompressZipFile( zipFilePath, destPath )
end
--[[

载入Lua打包文件

** Parameters: **

- filename:     string      lua包名称
- relativePath: string      相对res路径,设置则要以"/"结尾, 无则默认为""

--]]
function util.loadLuaZip( filename, relativePath )
    if ENABLED_LUA_ZIP_SOURCE then
        local relative = relativePath or ""
        local path = device.documentsPath .. relative .. filename
        if io.exists(path) then 
            CCLuaLoadChunksFromZip(path)
        else
            CCLuaLoadChunksFromZip(PATH_RES .. relative .. filename)
        end
    end
end
--=============================================================
--Convert to cpp obj
--=============================================================
function util.toCCNode( node ) return LuaNodeConverter:convertToCCNode(node) end
function util.toCCLayer( node ) return LuaNodeConverter:convertToCCLayer(node) end
function util.toCCSprite( node ) return LuaNodeConverter:convertToCCSprite(node) end
function util.toCCLabel( node ) return LuaNodeConverter:convertToCCLabelTTF(node) end
function util.toCCLabelAtlas( node ) return LuaNodeConverter:convertToCCLabelAtlas(node) end
function util.toCCLabelBMFont( node ) return LuaNodeConverter:convertToCCLabelBMFont(node) end
function util.toCCMenu( node ) return LuaNodeConverter:convertToCCMenu(node) end
function util.toCCMenuItem( node ) return LuaNodeConverter:convertToCCMenuItem(node) end
function util.toCCControlButton( node ) return LuaNodeConverter:convertToCCControlButton(node) end
function util.toCCTableView( node ) return LuaNodeConverter:convertToCCTableView(node) end
function util.toCCScrollView( node ) return LuaNodeConverter:convertToCCScrollView(node) end
function util.toCCScale9Sprite( node ) return LuaNodeConverter:convertToCCScale9Sprite(node) end
function util.toCCEditBox( node ) return LuaNodeConverter:convertToCCEditBox(node) end

function util.toXNode( node, nodeType )
    if not nodeType or not node then return node end 
    if nodeType == "CCNode" then 
        return util.toCCNode(node)
    elseif nodeType == "CCLayer" then 
        return util.toCCLayer( node )
    elseif nodeType == "CCSprite" then 
        return util.toCCSprite( node )
    elseif nodeType == "CCLabel" then 
        return util.toCCLabel( node )
    elseif nodeType == "CCLabelAtlas" then    
        return  util.toCCLabelAtlas( node )
    elseif nodeType == "CCLabelBMFont" then    
        return  util.toCCLabelBMFont( node )
    elseif nodeType == "CCMenu" then    
        return  util.toCCMenu( node )
    elseif nodeType == "CCMenuItem" then    
        return  util.toCCMenuItem( node )
    elseif nodeType == "CCControlButton" then    
        return  util.toCCControlButton( node )
    elseif nodeType == "CCTableView" then    
        return  util.toCCTableView( node )
    elseif nodeType == "CCScrollView" then    
        return  util.toCCScrollView( node )
    elseif nodeType == "CCScale9Sprite" then    
        return  util.toCCScale9Sprite( node )
    elseif nodeType == "CCEditBox" then    
        return  util.toCCEditBox( node )
    else
        return node
    end
end

--=========================================
--table扩展方法
--=========================================

function table.keyOfValue(t, value)
    if t and value and istable(t) then
        for k, v in pairs(t) do
            if v == value then return k end
        end
    end
    return nil
end

function table.slice( t, start, count )
    local result = {}
    if not istable(t) or not isnumber(start) then return result end 
    count = count or 0
    local endIndex = start + count - 1 
    if endIndex > #t then endIndex = #t end
    if t and start <= endIndex then
        for i = start, endIndex do
            result[#result + 1] = t[i]
        end 
    end
    return result
end

function table.contains( t, value )
    if table.isnotempty(t) then 
        for i, v in ipairs(t) do
            if v == value then 
                return true
            end
        end
    end
    return false
end

--[[ 

反转table

**Parameters:**

-   table:        table     原始table

**Returns:**

-   return 反转后table

--]]
function table.reverse( t )
    local result = {}
    local count = #t
    for i, v in ipairs(t) do
        result[i] = t[count + 1 - i]
    end
    return result
end
util.reversionTable = table.reverse

function table.isnotempty( t )
    local result = false
    if istable(t) then 
        for k,v in pairs(t) do
            result = true 
            break
        end
    end
    return result
end
--[[ 

从指定table中取出以指定属性的值为key,指定属性的值为value,抽取成新的table

**Parameters:**

-   t:              table       原始table
-   keyname:        string      做为key的属性名
-   valuename:      string      做为value的属性名

**Returns:**

-   return table 抽取结果

--]]
function table.pickupkv(t, keyname, valuename)
    local result = {}
    if table.isnotempty(t) then    
        for i, v in ipairs(t) do
            if istable(v) and v[keyname] and v[valuename] then 
                result[v[keyname]] = v[valuename]
            end
        end
    end 
    return result
end
--=========================================
--string扩展方法
--=========================================
--[[ 

判断字符串是否以特定字符串打头

**Parameters:**

-   str:        string      待判定字符串
-   startStr:   string      打头字符串
-   ignorCase:  bool        是否忽略大小写,默认为false

**Returns:**

-   return 字符串是否以特定字符串打头

--]]
function string.startwith( str, startStr, ignorCase )
    if isstring(str) and isstring(startStr) then 
        if ignorCase == true then 
            str = string.lower(str)
            startStr = string.lower(startStr)
        end
        return string.find(str, "^" .. startStr) ~= nil
    else
        return false
    end
end
--[[ 

判断字符串是否以特定字符串结尾

**Parameters:**

-   str:        string      待判定字符串
-   endStr:     string      结尾字符串
-   ignorCase:  bool        是否忽略大小写,默认为false

**Returns:**

-   return 字符串是否以特定字符串结尾

--]]
function string.endwith( str, endStr, ignorCase )
    if isstring(str) and isstring(endStr) then 
        if ignorCase == true then 
            str = string.lower(str)
            startStr = string.lower(startStr)
        end
        return string.find(str, endStr .. "$") ~= nil
    else
        return false
    end
end
--[[ 

将字符串首字母小写

**Parameters:**

-   str:        string      

**Returns:**

-   return 

--]]
function string.lcfirst( str )
    return string.lower(string.sub(str, 1, 1)) .. string.sub(str, 2)
end
--[[ 

将字符串以最后一个特定子串进行分割

**Parameters:**

-   str:        string     
-   sp:         string      分割字符串

**Returns:**

-   return 分割后的字符串前段, 后段

** e.g **

string.splitlast("/usr/local/ccb/init.lua", "/")
result:  "/usr/local/ccb", "init.lua"

--]]
function string.splitlast(str, sp)
    return string.match(str, string.format("(.+)%s(.+)", sp))
end
--[[ 

找到指定子串在字符串中的最后一个起始位置和结束位置

**Parameters:**

-   str:        string     
-   sp:         string      匹配的字符串

**Returns:**

-   return 起始位置, 结束位置       无时都nil, nil

** e.g **

string.findlast("/usr/local/ccb/init.lua", "/")
result: 15, 15

--]]
function string.findlast(str, subStr)
    local startIndex, length = nil, string.len(subStr)
    local i = 1
    while true do
        local index, len = string.find(str, subStr, i)
        if not index then 
            break
        else
            startIndex = index
            i = index + length
        end
    end
    local endIndex = nil
    if startIndex then endIndex = startIndex + length - 1 end
    return startIndex, endIndex
end

--=========================================
--LOG
--=========================================
local function createLog( logType, tag, fmt, ... )
    return string.format("[LUA:%s]%s %s", logType, tag, string.format(fmt, ...))
end
local function containsTag( tag )
    if not LOG_TAGS or not istable(LOG_TAGS) or #LOG_TAGS == 0 then
        return true 
    end
    for i,v in ipairs(LOG_TAGS) do
        if string.find(tag, v) then return true end
    end
    return false
end
--LOG_LEVEL: 0 = e; 1 = e,i; 2 = e,i,w; 3 = e,i,w,d; 4 = e,d.
if type(LOG_LEVEL) ~= "number" then LOG_LEVEL = 0 end
Log = {}
Log.d = function() end
Log.p = function () end
Log.d_CCArray  = function() end
Log.d_table  = function() end
Log.i  = function() end
Log.w  = function() end

if LOG_LEVEL >= 3 then
    Log.d = function ( TAG, fmt, ... )
        if containsTag( TAG ) then
            echo(createLog("DEBUG", TAG, fmt, ...))
        end
    end
    Log.p = function ( TAG, ... )
        if containsTag( TAG ) then
            print(TAG, ...)
        end
    end

    Log.d_CCArray = function ( TAG, ccarray, callback, fmt, ... )
        if containsTag( TAG ) then
            Log.d(TAG, fmt, ...)
            echo("============CCARRAY============")
            for i = 0, ccarray:count() - 1 do
                local obj = ccarray:objectAtIndex(i)
                local str
                if callback then
                    str = callback(obj)
                else
                    local node = util.toCCNode(obj)
                    str = node:getTag()
                end
                echo("["..i .. "]", str)
            end
            echo("===============================")
        end
    end

    Log.d_table = function ( TAG, t, callback, fmt, ... )
        if containsTag( TAG ) then
            Log.d(TAG, fmt, ...)
            echo("=============TABLE=============")
            for i,v in ipairs(t) do
                local str 
                if callback then
                    str = callback(v)
                else
                    str = v
                end
                echo("["..i.."]", str)
            end
            echo("===============================")
        end
    end
end

if LOG_LEVEL >= 1 and LOG_LEVEL ~= 4 then
    Log.i = function ( TAG, fmt, ... )
        if containsTag( TAG ) then
            echo(createLog("INFO", TAG, fmt, ...))
        end
    end
end

if LOG_LEVEL >= 2 and LOG_LEVEL ~= 4 then
    Log.w = function ( TAG, fmt, ... )
        if containsTag( TAG ) then
            echo(createLog("WARNING", TAG, fmt, ...))
        end
    end
end

Log.e = function ( TAG, fmt, ... )
   assert(nil, createLog("ERROR", TAG, fmt, ...))
end

Log.a = function ( cond, TAG, fmt, ... )
    assert(cond, createLog("ERROR", TAG, fmt, ...))
end
--[[ 

参数验证

**Parameters:**

-   TAG:        string      log输出用TAG
-   methodName: string      方法名,只是为了更好定位到哪个方法而已
-   params:     table       验证的参数,格式为{ {条件, 出错信息}, {条件, 出错信息}, ... }

--]]
function util.checkParams( TAG, methodName, params )
    for i = 1, #params do
        local p = params[i]
        local cond, info = p[1], p[2]
        if not cond then 
            if not info then 
                info = " invalid params"
            else
                info = " invalid params " .. info
            end
            info = methodName .. info
            Log.a(cond, TAG, info)
        end
    end
end
--=========================================
--不在util table中的util方法,直接调用
--=========================================

function iff( cond, trueValue, falseValue )
    if cond then
        return trueValue
    else
        return falseValue
    end
end

function ifnil( v, default )
    if isnil(v) then
        return default
    else
        return v
    end
end

function CCSizeAdd( size, width, height )
    return CCSize(size.width + width, size.height + height)
end

function CCSizeMul( size, width, height )
    return CCSize(size.width * width, size.height * height)
end
--[[ 

取得指定widget的类型       

**Parameters:**

-   w:        widget node     

**Returns:**

-   return string w的类型,无则为nil

--]]
function wtypeof( w )
    if w and w._type_ then
        return w._type_
    end
    return nil
end

function isnumber( v )
    return type(v) == "number"
end

function isstring( v )
    return type(v) == "string"
end

function isboolean( v )
    return type(v) == "boolean"
end

function istable( v )
    return type(v) == "table"
end

function isfunction( v )
    return type(v) == "function"
end

function isnil( v )
    return v == nil
end

function isnotnil( v )
    return v ~= nil
end

function isandroid()
    return device.platform == "android"
end

function isios()
    return device.platform == "ios"
end
--=============================================================
--多语言
--=============================================================
util.langStrings = {}

--[[ 

加载多语言文件,若未加载成功,将加载default

**Parameters:**

-   fileRequirePath:        string          导入多语言lua文件路径的前缀,如"language.lan_game",之后会自动加入手机语言,如"_cn"
-   language:               string          指定用哪种语言,如cn,en等
-   default:                string          最后无奈的时候用的语言, 默认为"_default"

** e.g. **
1.英文环境下
util.loadLanguage("language.lan_game")          ==> load language/lan_game_en.lua
2.强制指定使用中文
util.loadLanguage("language.lan_game", "cn")    ==> load language/lan_game_cn.lua

--]]
function util.loadLanguage( fileRequirePath, language, default )
    if fileRequirePath then 
        language = language or device.language
        xpcall(
            function( ... )
                util.langStrings = require(fileRequirePath .. "_" .. language)
            end,
            function( ... )
                util.langStrings = require(fileRequirePath .. (default or "_default"))
            end
        )
    end
end
--[[ 

取得指定文件的完整路径

**Parameters:**

-   fileName:        string     相对路径名

**Returns:**

-   return return

--]]
function util.fullPath( fileName )
    return CCFileUtils:sharedFileUtils():fullPathForFilename(fileName)
end

--[[ 

取得指定key多语言文字,若不存在且无默认值则直接返回key

**Parameters:**

-   key:        string      语言key
-   ...:        string      可替换的值

**Returns:**

-   return key对应的多语言信息

--]]
function _s(key, ...)
    -- return util.langStrings[key] or default or key
    local result = util.langStrings[key] or key
    if not ... then 
        return result
    else
        return string.format(result, ...)
    end
end
--[[ 

随机取得指定table中的指定数量元素

**Parameters:**

-   t:          table               待随机源
-   num:        int                 随机数量,默认为1

**Returns:**

-   return num为1,则直接返回随机到的元素,其它都返回结果table

**e.g.**
math.randomset( {-1, 1}, 1 ) ==> -1 or 1

--]]
function math.randomset( t, num )
    num = num or 1
    local result = t
    if istable(t) and isnumber(num) and #t ~= num then 
        result = {}
        if #t > num then
            local tt = clone(t)
            local count = #tt + 1
            for i = 1, num do
                local index = math.random(1, count - i )
                table.insert(result, tt[index])
                table.remove(tt, index)
            end
        elseif #t < num then
            local count = #t
            for i = 1, num do
                table.insert(result, t[math.random(1, count)])
            end
        end
    end
    if num == 1 and result and #result == 1 then 
        result = result[1]
    end
    return result
end

--=============================================================
--仿python内置函数
--=============================================================

--[[ 

将所有数据依次当作fun中的参数,返回其中结果为true的元素
若设置参数first,则是从过滤元素中找出前first个符合条件的元素返回

**Parameters:**

-   fun:        function(v,k) 过滤方法
-   params:     table         待过滤的元素
-   first:      int           只过滤指定数量的元素,不设置或设置小于1的数则都为全过滤

**Returns:**

-   return 过滤后的结果table

**e.g.**
util.filter(function(v) return v % 2 == 0 end , {1, 2, 3, 4, 5})  ==> {2, 4}

--]]
function util.filter(fun, params, first)
    local result = {}
    if isfunction(fun) and istable(params) then 
        local index = 1
        if not (isnumber(first) and first > 0) then 
            first = nil
        end
        for k, v in pairs(params) do
            if fun(v, k) then 
                result[index] = v
                if first and index >= first then 
                    break
                end
                index = index + 1
            end
        end
    end
    return result
end
--[[ 

将所有值依次当作fun的参数,并将所有结果以table形式返回

**Parameters:**

-   fun:        function(v,k) 方法
-   params:     table         待过滤的元素

**Returns:**

-   return table 所有结果

**e.g.**
util.map(function(t) 
            local a, b = unpack(t)
            return a + b
        end, {{1, 1}, {2, 2}, {3, 3}})
==> {2, 4, 6}
--]]
function util.map(fun, params)
    local result = {}
    if isfunction(fun) and istable(params) then 
        local index = 1
        for k, v in pairs(params) do
            result[index] = fun(v, k)
            index = index + 1
        end
    end
    return result
end
--[[ 

将所有值依次当作fun的参数执行一次,无返回结果

**Parameters:**

-   fun:        function(v, k)  方法
-   params:     table           待过滤的元素

**e.g.**
util.map(function(t) 
            print(t[1], t[2])
        end, {{1, 1}, {2, 2}, {3, 3}})
--]]
function util.walk(fun, params)
    if isfunction(fun) and istable(params) then 
        for k, v in pairs(params) do
            fun(v, k)
        end
    end
end
--[[ 

将params头两个当作参数给fun,并将结果及下一个元素当参数给fun,直到结束,返回结果

**Parameters:**

-   fun:        function      方法
-   params:     table         待过滤的元素

**Returns:**

-   return 根据实际情况

**e.g.**
util.reduce(function(a, b) return a .. b end , "a", "b", "c") ==> "abc"

--]]
function util.reduce(fun, params)
    local result
    if isfunction(fun) and istable(params) then 
        if #params >= 2 then 
            result = params[1]
            for i = 2, #params do
                result = fun(result, params[i])
            end
        end
    end
    return result
end
--[[ 

管道处理数据,...表示要处理的function管道,即多个function,每个都有返回值,data传入第一个
function并得到新的结果,再把新的结果带到下一个function,直至结束

**Parameters:**

-   data:       任意             初始参数
-   ...:        functions       需要经过的所有函数

**Returns:**

-   return 任意 结果

**e.g.**
local function even_filter(nums)
    return util.filter(function (v) return v % 2 == 0 end, nums)
end
local function multiply_by_three(nums)
    return util.map(function (v) return v * 3 end, nums)
end
local function convert_to_string(nums)
    util.walk(function (v) print("The Number:" .. v) end, nums)
end
util.pipeline(util.range(10), even_filter, multiply_by_three, convert_to_string)

--]]
function util.pipeline(data, ...)
    return util.reduce(function (data, fun)
        if data and isfunction(fun) then 
            return fun(data)
        else
            return data
        end
    end, {data, ...})
end
--[[ 

取得指定范围内的数据集

**Parameters:**

-   s:          int             起始值, 若无e,则此为结束值,起始值默认为1
-   e:          int             结束值
-   step:       int             步长,默认为1 

**Returns:**

-   return table 指定范围内的数据集

**e.g.**
util.range(3) ==> {1, 2, 3}
util.range(1, 4, 2) ==> {1, 3}

--]]
function util.range(s, e, step)
    local result = {}
    if s then 
        step = step or 1
        if not e then 
            s, e = 1, s
        end
        local index = 1
        for i = s, e, step do
            result[index] = i
            index = index + 1
        end
    end
    return result
end
--[[ 

是否所有参数都满足fun

**Parameters:**

-   fun:        function      方法,必须返回bool值
-   params:     table         待判定的元素

**Returns:**

-   return bool 是否满足

**e.g.**
print(util.all(function (v) return v % 2 == 0 end, {4,2,0})) ==> true

--]]
function util.all(fun, params)
    if not isfunction(fun) or not istable(params) then 
        return false
    end
    for i, v in ipairs(params) do
        if not fun(v) then 
            return false
        end
    end
    return true
end
--[[ 

是否存在至少一个参数满足fun

**Parameters:**

-   fun:        function      方法,必须返回bool值
-   params:     table         待判定的元素

**Returns:**

-   return bool 是否满足

**e.g.**
print(util.any(function (v) return v % 2 ~= 0 end, {4,2,1})) ==> true

--]]
function util.any(fun, params)
    if not isfunction(fun) or not istable(params) then 
        return false
    end
    for i, v in ipairs(params) do
        if fun(v) == true then 
            return true
        end
    end
    return false
end
--=============================================================
--通用颜色
--=============================================================
COLOR3 = {
    WHITE   = ccc3(255, 255, 255),
    BLACK   = ccc3(0, 0, 0),
    RED     = ccc3(255, 0, 0),
    GREEN   = ccc3(0, 255, 0),
    BLUE    = ccc3(0, 0, 255),
    YELLOW  = ccc3(255, 255, 0),
    GREE    = ccc3(166, 166, 166),
    ORANGE  = ccc3(255, 127, 0),
    MAGENTA = ccc3(255, 0, 255),
}
COLOR_MASK = ccc4(16, 8, 11, 153)
BUTTON_COLOR3 = {
    BROWN   = ccc3(96, 57, 19),
    RED     = ccc3(121, 4, 16),
    GREEN     = ccc3(0, 73, 35),
    BLUE     = ccc3(38, 34, 98),
    GREE     = ccc3(100, 100, 100),
}
--=============================================================
--通用0值
--=============================================================
CCPointZero = CCPointMake(0,0);
CCSizeZero = CCSizeMake(0,0);
CCRectZero = CCRectMake(0,0,0,0);

return util