--[[

Copyright (c) 2012-2013 gao7.com

http://www.gao7.com/LizardMan/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

]]

--[[--

抽象动作类，定义公共动作相关操作方法及逻辑实现。

-   定义基础动作方法及成员变量。

]]

local M = class("AbstractAction")



-- [包含文件]
-- ..


-- [操作变量]
-- ..


-- [公共参数]
-- [常量]
-- ..

    

    





-- [构造方法]
-- 构造方法，定义视图加载逻辑
function M:ctor(params)
    -- [合法性验证]
    self:assertParameters(params)



    -- [参数赋值]
    -- [结点]
    self._node                  	= params.node                                           -- 结点
    -- [动作模板]
    self._isStopAllActions      	= params.isStopAllActions   		or true          	-- 标识是否停止所有动作
    self._delayTimeBegin            = params.delayTimeBegin             or 0             	-- 延迟时间[起始]
    self._delayTimeEnd              = params.delayTimeEnd               or 0                -- 延迟时间[结束]
    self._callfront                 = params.callfront                                      -- 前调函数
    self._callfrontParams           = params.callfrontParams            or {}               -- 前调函数参数
    self._callback              	= params.callback                                       -- 回调函数
    self._callbackParams        	= params.callbackParams     		or {}              	-- 回调函数参数
    self._isReassignPositionBegin   = params.isReassignPositionBegin 	or false          	-- 是否重新指派位置[起始]
    self._positionBegin      		= params.positionBegin      		or P_ZERO		    -- 动作起始位置
    self._isReassignPositionEnd    	= params.isReassignPositionEnd 		or false         	-- 是否重新指派位置[结束]
    self._positionEnd      			= params.positionEnd      			or P_ZERO		    -- 动作终止位置
    -- [标识]
    self._sync                      = params.sync                       or SYNC_IGNORE      -- 同步标识
    -- [代理]
    self._delegate                  = params.delegate                                       -- 代理
    -- [后续动作]
    self._successors         		= params.successors                                     -- 后继者集



    -- [具体操作]
    -- 预加载
    self:preload(params)
end

    

    





-- [抽象方法]
-- 构建动作对象
function M:buildAction(params)
	TDExceptionUtils.raiseUnimpletentException("AbstractFightAction:buildAction()")
end

    

    





-- [模板方法]
-- 运行动作
function M:run(params)
    -- [代理逻辑]
    if self._delegate then 
        self._delegate:delActionRun(self, params)
    end
    -- print("i:" .. tostring(self._index) .. "   class: ", self.__cname)

-- print("开始运行啦。。。。。。。。。。。。")



    -- [动作逻辑]
    -- 动画参数
    local target            = self._node                                -- 目标对象
    local timeDelayBegin    = self:T(self._delayTimeBegin)              -- 延迟时间[起始]
    local timeDelayEnd      = self:T(self._delayTimeEnd)                -- 延迟时间[结束]



	-- 运行动作前操作
    self:hookRunStart(params)



    -- 初始配置
    -- 是否停止动作序列
    if self._isStopAllActions then 
        target:stopAllActions()
    end
    -- 位置还原[起始位置]
   	if self:isReassignPositionBegin() then
   		target:setPosition(self:getPositionBegin())
   	end
    






    -- 函数定义
    -- 前调函数
    local function onStart()
        -- 处理回调逻辑
        local callfront         = self._callfront
        local callfrontParams   = self._callfrontParams 

        if type(callfront) ~= "function" then 
            callfront = nil 
        end 
        if callfront then
            callfront(callfrontParams)
        end
    end
    -- 回调函数
    local function onComplete()
        -- 处理回调逻辑
        local callback          = self._callback
        local callbackParams    = self._callbackParams 

        if type(callback) ~= "function" then 
            callback = nil 
        end 
        if callback then
            callback(callbackParams)
        end

        
    	-- 位置还原[终止位置]
        if self:isReassignPositionEnd() then
            target:setPosition(self:getPositionEnd())
        end


		-- 运行动作后操作
	    self:hookRunEnd(params)
    end


    -- 动作序列
    local actions = {}

    -- 1、添加动作[延迟]
    if timeDelayBegin > 0 then
        actions[#actions + 1] = CCDelayTime:create(timeDelayBegin)
    end

    -- 2、添加动作[回调函数]
    if self._callfront then
        actions[#actions + 1] = CCCallFunc:create(onStart)
    end
    
    -- 3、添加动作[构建动作]
    actions[#actions + 1] = self:buildAction(params)

    -- 4、添加动作[回调函数]
    actions[#actions + 1] = CCCallFunc:create(onComplete)

    -- 5、添加动作[延迟]
    if timeDelayEnd > 0 then
        actions[#actions + 1] = CCDelayTime:create(timeDelayEnd)
    end

    -- 6、添加动作[处理后继者逻辑]
    local actionSuccessor = self:getSuccessorAction()
    if actionSuccessor then 
        actions[#actions + 1] = self:getSuccessorAction()
    end

    -- 运行动作
    target:runAction(transition.sequence(actions))
end

-- 停止动作
function M:stop()
    self._node:stopAllActions()
end

    

    





-- [勾子方法]
-- 运行动作前操作
function M:hookRunStart(params)

end

-- 运行动作后操作
function M:hookRunEnd(params)
	
end

    

    





-- [辅助方法]
-- 获得动作时间
function M:T(time)
	return FightActionHelper:time(time)
end

-- 获得位置
function M:P(position)
	return FightActionHelper:position(position)
end

-- 获得位置偏移
function M:O(position, offset)
    return FightActionHelper:offset(position, offset)
end

-- 获得位置
function M:D(distance)
    return FightActionHelper:distance(distance)
end

-- 获得旋转角度
function M:R(rotation)
	return FightActionHelper:rotation(rotation)
end

-- 获得缩放比例
function M:S(scale)
    return FightActionHelper:scale(scale)
end

-- 获得透明度
function M:A(alpha)
    return FightActionHelper:alpha(alpha)
end

-- 获得结点
function M:N(node)
	return FightActionHelper:node(node)
end




    





-- [实例方法]
-- [实例方法][实体渲染]
-- 实体渲染
function M:entityRender()

end


---------


-- [实例方法][实体析构]
-- 实体析构
function M:entityDestructor()

end


---------


-- [实例方法][预加载]
-- 预加载
function M:preload(params)
    -- 预加载[资源]
    self:preloadResource(params)
end

-- 预加载[资源]
function M:preloadResource(params)
    local images = self:getResourcesImage(params)             -- 图片
    local sounds = self:getResourcesSound(params)             -- 声音

    if TDArrayUtils.isNotEmpty(images) then 
        for _, resourceName in pairs(images) do
             -- print("文件名称：：：：：："..self:pathResource(resourceName))
            TDAnimationUtils.addSpriteFramesWithResource(self:pathResource(resourceName))
        end
    end
    if TDArrayUtils.isNotEmpty(sounds) then 
        for _, resourceName in pairs(sounds) do
            SOUND.preloadSound(self:pathSound(resourceName))
        end
    end
end

-- 获得路径[资源文件]
function M:pathResource(resourceName)
    return resourceName
end

-- 获得路径[音乐文件]
function M:pathSound(resourceName)
    return resourceName
end

-- 获得预加载的资源[图片]
function M:getResourcesImage(params)
    return nil
end

-- 获得预加载的资源[音效]
function M:getResourcesSound(params)
    return nil
end


---------


-- [实例方法][外部调用]
-- 获得后继者动作（Action）
function M:getSuccessorAction()
	-- 返回值
	local action = nil

	-- 获得后继者集
	local successors = self:getSuccessors()
	-- 非空判断
	if TDArrayUtils.isNotEmpty(successors) then
		action =  
			CCCallFunc:create(
				function()
                    local successors = self:getSuccessors()
                    for _, successor in ipairs(successors) do
                        successor:run()
                    end
				end
			)
	end

	return action
end

-- 获得类名
function M:getClassName()
    return self.__cname
end


---------


-- [实例方法][验证]
-- 验证合法性[构造函数参数]
function M:assertParameters(params)
	-- 验证参数集
    TDAssertUtils.assertParameters(params)
end


---------


-- [实例方法][属性Getter/Setter]
-- [结点]
-- 获得结点
function M:getNode()
    return self._node
end
-- 设置结点
function M:setNode(node)
    self._node = node
end


---------


-- [实例方法][动作模板]
-- 获得标识是否停止所有动作
function M:isStopAllActions()
    return self._node
end
-- 设置标识是否停止所有动作
function M:setStopAllActions(isStopAllActions)
    self._isStopAllActions = isStopAllActions
end

-- 获得延迟时间[起始]
function M:getDelayTimeBegin()
    return self._delayTimeBegin
end
-- 设置延迟时间[起始]
function M:setDelayTimeBegin(delayTime)
    self._delayTimeBegin = delayTime
end

-- 获得延迟时间[结束]
function M:getDelayTimeEnd()
    return self._delayTimeEnd
end
-- 设置延迟时间[结束]
function M:setDelayTimeEnd(delayTime)
    self._delayTimeEnd = delayTime
end

-- 获得前调函数
function M:getCallfront()
    return self._callfront
end
-- 设置前调函数
function M:setCallfront(callfront)
    self._callfront = callfront
end

-- 获得前调函数参数
function M:getCallfrontParams()
    return self._callfrontParams
end
-- 设置前调函数参数
function M:setCallfrontParams(callfrontParams)
    self._callfrontParams = callfrontParams
end

-- 获得回调函数
function M:getCallback()
    return self._callback
end
-- 设置回调函数
function M:setCallback(callback)
    self._callback = callback
end

-- 获得回调函数参数
function M:getCallbackParams()
    return self._callbackParams
end
-- 设置回调函数参数
function M:setCallbackParams(callbackParams)
    self._callbackParams = callbackParams
end

-- 获得是否重新指派[起始位置]
function M:isReassignPositionBegin()
    return self._isReassignPositionBegin
end
-- 设置是否重新指派[起始位置]
function M:setReassignPositionBegin(isReassign)
    self._isReassignPositionBegin = isReassign
end

-- 获得动作起始位置
function M:getPositionBegin()
    return self._positionBegin
end
-- 设置动作起始位置
function M:setPositionBegin(position)
    self._positionBegin = position
end

-- 获得动作终止位置
function M:getPositionEnd()
    return self._positionEnd
end
-- 设置动作终止位置
function M:setPositionEnd(position)
    self._positionEnd = position
end

-- 获得是否重新指派[终止位置]
function M:isReassignPositionEnd()
    return self._isReassignPositionEnd
end
-- 设置是否重新指派[终止位置]
function M:setReassignPositionEnd(isReassign)
    self._isReassignPositionEnd = isReassign
end


---------


-- [后续动作]
-- 获得后继者
function M:getSuccessor()
    TDAssertUtils.assertNotNil(self._successors)
    TDAssertUtils.assertNotNil(self._successors[1])

    return self._successors[1]
end
-- 设置后继者
function M:setSuccessor(successor)
    self._successors = { successor }
end

-- 获得后继者集
function M:getSuccessors()
    return self._successors
end
-- 设置后继者集
function M:setSuccessors(successors)
    self._successors = successors
end


---------


-- [同步]
-- 获得同步标识
function M:getSync()
    return self._sync
end





return M