--[[

Copyright (c) 2012-2013 gao7.com

http://www.gao7.com/LizardMan/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

]]

--[[--

字符串工具类，定义字符串相关操作方法及逻辑实现。

-   指定字符串逻辑。
-   公共字符串定义。

]]

local M = {}



-- [公共参数]
-- 常量

-- 操作变量
-- 断言
local TDAssertUtils      			= require(PH_MDL_UTIL .. "TDAssertUtils")





-- [私有方法]
-- [声明]

-- [定义]





-- [公有方法]
--[[--

字符串截取

### Description:

    TDStringUtils.split(str, delim, maxNb)

### Example:
    local str    	= "a|b|c|d"
    local charArr 	= TDStringUtils.split(str, "|")
	dump(charArr)                                       --> { "a", "b", "c", "d" }

    charArr         = TDStringUtils.split(str, "|", 2) 
    dump(charArr)                                       --> { "a", "b" }
]]
function M.split(str, delim, maxNb)
    -- Eliminate bad cases...   
    if string.find(str, delim) == nil then  
        return { str }  
    end
    if maxNb == nil or maxNb < 1 then  
        maxNb = 0    -- No limit   
    end

    local result = {}  
    local pat = "(.-)" .. delim .. "()"   
    local nb = 0  
    local lastPos   
    for part, pos in string.gfind(str, pat) do  
        nb = nb + 1  
        result[nb] = part   
        lastPos = pos   
        if nb == maxNb then break end  
    end  
    -- Handle the last field
    if nb ~= maxNb then  
        result[nb + 1] = string.sub(str, lastPos)   
    end

    return result
end

--[[--

将字符串拆解成字符数组

### Description:

    TDStringUtils.split(str, maxLength)

### Example:
    local str       = "abcd"
    local charArr   = TDStringUtils.toCharArray(str)
    dump(charArr)                                       --> { "a", "b", "c", "d" }

    charArr         = TDStringUtils.toCharArray(str, 3) 
    dump(charArr)                                       --> { "a", "b" }
]]
function M.toCharArray(str, maxLength)
    TDAssertUtils.assert(str ~= nil)
    TDAssertUtils.assert(maxLength == nil or (type(maxLength) == "number" and maxLength > 0))

    local array = {}
    local length = maxLength or M.length(str)
    -- print(length)
    for i = 1, length do
        array[#array + 1] = string.sub(str, i, i)
    end

    return array
end

--[[--

获得字符串长度

### Description:

    TDStringUtils.length(str)

### Example:
    local str       = "a|b|c|d"
    local length    = TDStringUtils.length(str)
    print("length: ", length)               --> length:  7
]]
function M.length(str)
    return string.len(str)
end

--[[--

返回重复2次字符串s的串

### Description:

    TDStringUtils.repeat(str, num)

### Example:
    local str       = "team"
    print(TDStringUtils.double(str))        --> teamteam
]]
function M.double(str)
    return M.ditto(str, 2)
end

--[[--

返回重复n次字符串s的串；你使用string.rep("a", 2^20)可以创建一个1M bytes的字符串（比如，为了测试需要）
说明：
    因为TDStringUtils.repeat中repeat为关键词，不能使用，故使用ditto代替

### Description:

    TDStringUtils.ditto(str, num)

### Example:
    local str       = "team"
    print(TDStringUtils.ditto(str))         --> teamteam
    print(TDStringUtils.ditto(str, 4))      --> teamteamteamteam
]]
function M.ditto(str, num)
    return string.rep(str, num or 2)
end

--[[--

将字符串中的小写字母转换成大写

string.lower(s)       将s中的大写字母转换成小写（string.upper将小写转换成大写）。如果你想不关心大小写对一个数组进行排序的话，你可以这样：
                             table.sort(a, function (a, b) return string.lower(a) < string.lower(b) end)

### Description:

    TDStringUtils.toUpperCase(str)

### Example:
    local str       = "team"
    print(TDStringUtils.toUpperCase(str))       --> TEAM
]]
function M.toUpperCase(str)
    return string.upper(str)
end

--[[--

将字符串中的大写字母转换成小写

说明：
    如果你想不关心大小写对一个数组进行排序的话，你可以这样：
    table.sort(a, function (a, b) return string.lower(a) < string.lower(b) end)

### Description:

    TDStringUtils.toLowerCase(str)

### Example:
    local str       = "TEAM"
    print(TDStringUtils.toLowerCase(str))       --> team
]]
function M.toLowerCase(str)
    return string.lower(str)
end

--[[--

格式化字符串

### Description:

    TDStringUtils.format(...)
    生成具有特定格式的字符串, 函数的第一个参数是格式(formatstring), 之后是对应格式中每个代号的各种数据. 由于格式字符串的存在, 使得产生的长字符串可读性大大提高了. 

    格式字符串可能包含以下的转义码:
        %c - 接受一个数字, 并将其转化为ASCII码表中对应的字符
        %d, %i - 接受一个数字并将其转化为有符号的整数格式
        %o - 接受一个数字并将其转化为八进制数格式
        %u - 接受一个数字并将其转化为无符号整数格式
        %x - 接受一个数字并将其转化为十六进制数格式, 使用小写字母
        %X - 接受一个数字并将其转化为十六进制数格式, 使用大写字母
        %e - 接受一个数字并将其转化为科学记数法格式, 使用小写字母e
        %E - 接受一个数字并将其转化为科学记数法格式, 使用大写字母E
        %f - 接受一个数字并将其转化为浮点数格式
        %g(%G) - 接受一个数字并将其转化为%e(%E, 对应%G)及%f中较短的一种格式
        %q - 接受一个字符串并将其转化为可安全被Lua编译器读入的格式
        %s - 接受一个字符串并按照给定的参数格式化该字符串

### Example:

    print(TDStringUtils.format("pi = %.4f", 3.1415926))         --> pi = 3.1416
    d = 5; m = 11; y = 1990
    print(TDStringUtils.format("%02d/%02d/%04d", d, m, y))      --> 05/11/1990   
    tag, title = "h1", "a title"
    print(TDStringUtils.format("<%s>%s</%s>", tag, title, tag)) --> <h1>a title</h1>

    TDStringUtils.format("%%c: %c", 83)                         --> S
    TDStringUtils.format("%+d", 17.0)                           --> +17
    TDStringUtils.format("%05d", 17)                            --> 00017
    TDStringUtils.format("%o", 17)                              --> 21
    TDStringUtils.format("%u", 3.14)                            --> 3
    TDStringUtils.format("%x", 13)                              --> d
    TDStringUtils.format("%X", 13)                              --> D
    TDStringUtils.format("%e", 1000)                            --> 1.000000e+03
    TDStringUtils.format("%E", 1000)                            --> 1.000000E+03
    TDStringUtils.format("%6.3f", 13)                           --> 13.000
    TDStringUtils.format("%q", "One\nTwo")                      --> "One\
                                            　　Two"
    TDStringUtils.format("%s", "monkey")                        --> monkey
    TDStringUtils.format("%10s", "monkey")                      -->     monkey
    TDStringUtils.format("%5.3s", "monkey")                     -->   mon
]]
function M.format(...)
    return string.format(...)
end





return M